"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 7, 2012.
"""
from copy import deepcopy
from everest.representers.converters import ConverterRegistry
from everest.representers.converters import SimpleConverterRegistry
from everest.representers.interfaces import IRepresentationConverter
from everest.rfc3339 import rfc3339
from everest.testing import Pep8CompliantTestCase
from iso8601.iso8601 import parse_date
from pytz import timezone
from zope.interface import provider # pylint: disable=E0611,F0401
import datetime

__docformat__ = 'reStructuredText en'
__all__ = ['RepresenterConverterTestCase',
           ]


class RepresenterConverterTestCase(Pep8CompliantTestCase):
    def test_registry(self):
        with self.assert_raises(ValueError) as cm:
            SimpleConverterRegistry.register(bool, MyConverter)
        self.assert_true(str(cm.exception).startswith('For %s' % bool))
        with self.assert_raises(ValueError) as cm:
            SimpleConverterRegistry.register(MyNumberType, MyInvalidConverter)
        self.assert_true(str(cm.exception).startswith('Converter class'))

    def test_default_converter(self):
        self.assert_equal(
                MyConverterRegistry.convert_from_representation('0', int), 0)
        setattr(MyConverterRegistry, '_ConverterRegistry__converters', None)
        self.assert_equal(
                MyConverterRegistry.convert_to_representation(0, int), '0')
        setattr(MyConverterRegistry, '_ConverterRegistry__converters', None)

    def test_bool_converter(self):
        self.assert_true(
            SimpleConverterRegistry.convert_from_representation('true', bool)
            is True)
        self.assert_true(
            SimpleConverterRegistry.convert_from_representation(None, bool)
            is None)
        self.assert_equal(
            SimpleConverterRegistry.convert_to_representation(True, bool),
            'true')

    def test_datetime_converter(self):
        utc = timezone('UTC')
        ldt = datetime.datetime(2012, 8, 29, 16, 20, 0, tzinfo=utc)
        ldt_rpr = rfc3339(ldt, use_system_timezone=False)
        self.assert_equal(
            SimpleConverterRegistry.convert_from_representation(
                                                        ldt_rpr,
                                                        datetime.datetime),
            ldt)
        self.assert_equal(
            SimpleConverterRegistry.convert_to_representation(ldt,
                                                        datetime.datetime),
            ldt_rpr)

    def test_parse_date_fix(self):
        d = parse_date('2012-06-13 11:06:47+02:00')
        d_copy = deepcopy(d)
        self.assert_equal(d, d_copy)


class MyNumberType(int):
    pass


class MyInvalidConverter(object):
    pass


@provider(IRepresentationConverter)
class MyConverter(object):
    @classmethod
    def convert_from_representation(cls, representation_value,
                                    value_type): # pylint: disable=W0613
        return representation_value

    @classmethod
    def convert_to_representation(cls, value,
                                  value_type): # pylint: disable=W0613
        return value


class MyConverterRegistry(ConverterRegistry):
    pass



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from everest.querying.operators import ASCENDING
from everest.querying.operators import DESCENDING
from everest.querying.orderparser import parse_order
from everest.querying.specifications import AscendingOrderSpecification
from everest.querying.specifications import ConjunctionOrderSpecification
from everest.querying.specifications import DescendingOrderSpecification
from everest.testing import TestCaseWithConfiguration
from pyparsing import ParseException

__docformat__ = 'reStructuredText en'
__all__ = ['OrderParserTestCase',
           ]


class OrderParserTestCase(TestCaseWithConfiguration):
    def set_up(self):
        TestCaseWithConfiguration.set_up(self)
        self.parser = parse_order

    def test_no_criterion_query(self):
        expr = ''
        self.assert_raises(ParseException, self.parser, expr)

    def test_one_sort_order(self):
        expr = 'name:asc'
        result = self.parser(expr)
        self.assert_true(isinstance(result, AscendingOrderSpecification))
        self.assert_equal(result.attr_name, 'name')
        self.assert_equal(result.operator, ASCENDING)

    def test_one_sort_order_reversed(self):
        expr = 'name:desc'
        result = self.parser(expr)
        self.assert_true(isinstance(result, DescendingOrderSpecification))
        self.assert_equal(result.attr_name, 'name')
        self.assert_equal(result.operator, DESCENDING)

    def test_two_sort_order_left_reversed(self):
        expr = 'name:desc~age:asc'
        result = self.parser(expr)
        self.assert_true(isinstance(result, ConjunctionOrderSpecification))
        self.assert_true(isinstance(result.left,
                                    DescendingOrderSpecification))
        self.assert_true(isinstance(result.right,
                                    AscendingOrderSpecification))
        self.assert_equal(result.left.attr_name, 'name')
        self.assert_equal(result.left.operator, DESCENDING)
        self.assert_equal(result.right.attr_name, 'age')
        self.assert_equal(result.right.operator, ASCENDING)



"""
Repository base classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 5, 2013.
"""
from everest.entities.utils import get_entity_class
from everest.repositories.interfaces import IRepository
from everest.resources.base import Collection
from everest.resources.utils import get_collection_class
from zope.interface import implementer # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['AutocommittingSessionMixin',
           'Repository',
           'Session',
           'SessionFactory',
           ]


class SessionFactory(object):
    """
    Base class for session factories.
    """
    def __init__(self, repository):
        self._repository = repository

    def __call__(self):
        raise NotImplementedError('Abstract method.')


class Session(object):
    """
    Abstract base class for session objects.
    """
    #: Flag to indicate whether session operations need to update back
    #: references.
    IS_MANAGING_BACKREFERENCES = None

    def get_by_id(self, entity_class, id_key):
        """
        Retrieves the entity for the specified entity class and ID.

        :param entity_class: the type of the entity to retrieve.
        :param id_key: ID of the entity to retrieve.
        """
        raise NotImplementedError('Abstract method.')

    def get_by_slug(self, entity_class, slug):
        """
        Retrieves the entity for the specified entity class and slug.

        :param entity_class: the type of the entity to retrieve.
        :param slug: slug of the entity to retrieve.
        """
        raise NotImplementedError('Abstract method.')

    def add(self, entity_class, data):
        """
        Adds the given entity of the given entity class to the session.

        At the point an entity is added, it must not have an ID or a slug
        of another entity that is already in the session. However, both the ID
        and the slug may be ``None`` values.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter` or an
          iterable of such objects.
        """
        raise NotImplementedError('Abstract method.')

    def remove(self, entity_class, data):
        """
        Removes the specified of the given entity class from the session.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter` or an
          iterable of such objects.
        :raises ValueError: If the entity data does not provide an ID
            (unless it is marked NEW).
        """
        raise NotImplementedError('Abstract method.')

    def update(self, entity_class, data, target=None):
        """
        Updates an existing entity with the given entity data. If
        :param:`target_data` not given, the target entity will be determined
        through the ID supplied with the data.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter` or an
          iterable of such objects.
        :raises ValueError: If no target is given and the session does not
          contain an entity with the ID provided with the data.
        """
        raise NotImplementedError('Abstract method.')

    def query(self, entity_class):
        raise NotImplementedError('Abstract method.')

    def begin(self):
        raise NotImplementedError('Abstract method.')

    def commit(self):
        raise NotImplementedError('Abstract method.')

    def rollback(self):
        raise NotImplementedError('Abstract method.')

    def reset(self):
        raise NotImplementedError('Abstract method.')


class AutocommittingSessionMixin(object):
    """
    Mixin classes for sessions that wrap every add, remove, and update
    operation into a transaction.
    """
    def add(self, entity_class, data):
        self.begin()
        super(AutocommittingSessionMixin, self).add(entity_class, data)
        self.commit()

    def remove(self, entity_class, data):
        self.begin()
        super(AutocommittingSessionMixin, self).remove(entity_class, data)
        self.commit()

    def update(self, entity_class, data, target=None):
        self.begin()
        spr = super(AutocommittingSessionMixin, self)
        updated_entity = spr.update(entity_class, data, target=target)
        self.commit()
        return updated_entity


@implementer(IRepository)
class Repository(object):
    """
    Base class for repositories.

    A repository has the following responsibilities:
     * Configure and initialize a storage backend for resource data;
     * Create and cache aggregate and collection accessors for registered
       resources;
     * Create and hold a session factory which is used to create a
       (thread-local) session. The session is used by the accessors to
       load entities and resources from the repository.
    """

    #: A list of key names which can be used by :method:`configure`.
    _configurables = ['messaging_enable', 'messaging_reset_on_start']

    def __init__(self, name, aggregate_class,
                 join_transaction=False, autocommit=False):
        """
        Constructor.

        :param name: Name for this repository (propagated to repository).
        :param aggregate_class: The aggregate class to use when creating new
          aggregates in this repository.
        :param join_transaction: Indicates whether this repository should
          participate in the Zope transaction.
        :param autocommit: Indicates whether changes should be committed
          automatically.
        """
        self.__name = name
        if join_transaction and autocommit:
            raise ValueError('The "join_transaction" flag and the '
                             '"autocommit" flag can not both be set.')
        #: Flag indicating that the sessions using this repository should
        #: join the Zope transaction.
        self.join_transaction = join_transaction
        #: Flag indicating that changes should be committed immediately.
        self.autocommit = autocommit
        self._config = {}
        self.__is_initialized = False
        self.__cache = {}
        self.__agg_cls = aggregate_class
        self.__session_factory = None
        #: The set of resources (collection classes) managed by this
        #: repository.
        self.__registered_resources = set()

    def get_aggregate(self, resource):
        return self.get_collection(resource).get_aggregate()

    def get_collection(self, resource):
        if not self.__is_initialized:
            raise RuntimeError('Repository needs to be initialized.')
        ent_cls = get_entity_class(resource)
        root_coll = self.__cache.get(ent_cls)
        if root_coll is None:
            # Create a new root aggregate.
            root_agg = self.__agg_cls.create(ent_cls, self.session_factory,
                                             self)
            # Create a new root collection.
            coll_cls = get_collection_class(resource)
            root_coll = coll_cls.create_from_aggregate(root_agg)
            self.__cache[ent_cls] = root_coll
        return root_coll.clone()

    def set_collection_parent(self, resource, parent):
        """
        Sets the parent of the specified root collection to the given
        object (typically a service object).

        :param resource: Registered resource.
        :raises ValueError: If no root collection has been created for the
          given registered resource.
        """
        ent_cls = get_entity_class(resource)
        root_coll = self.__cache.get(ent_cls)
        if root_coll is None:
            raise ValueError('No root collection available for resource.')
        root_coll.__parent__ = parent

    def configure(self, **config):
        for key, val in config.items():
            if not key in self._configurables:
                raise ValueError('Invalid configuration key "%s".' % key)
            self._config[key] = val

    @property
    def configuration(self):
        return self._config.copy()

    def initialize(self):
        self._initialize()
        self.__is_initialized = True

    @property
    def is_initialized(self):
        return self.__is_initialized

    def register_resource(self, resource):
        if not issubclass(resource, Collection):
            resource = get_collection_class(resource)
        self.__registered_resources.add(resource)

    @property
    def registered_resources(self):
        return iter(self.__registered_resources)

    def is_registered_resource(self, resource):
        return get_collection_class(resource) in self.__registered_resources

    @property
    def session_factory(self):
        if self.__session_factory is None:
            # Create a session factory. The call to _initialize may depend on
            # the session factory being instantiated.
            self.__session_factory = self._make_session_factory()
        return self.__session_factory

    @property
    def name(self):
        return self.__name

    def reset(self):
        if not self.__session_factory is None:
            self.__session_factory().reset()
        self.__cache.clear()
        self._reset()
        self.__is_initialized = False

    def _initialize(self):
        """
        Performs initialization of the repository.
        """
        raise NotImplementedError('Abstract method.')

    def _reset(self):
        """
        Performs reset of the repository.
        """
        raise NotImplementedError('Abstract method.')

    def _make_session_factory(self):
        """
        Create the session factory for this repository.
        """
        raise NotImplementedError('Abstract method.')



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from everest.entities.interfaces import IEntity
from everest.repositories.rdb.utils import as_slug_expression
from everest.repositories.rdb.utils import get_metadata
from everest.repositories.rdb.utils import hybrid_descriptor
from everest.repositories.rdb.utils import is_metadata_initialized
from everest.repositories.rdb.utils import mapper
from everest.repositories.rdb.utils import reset_metadata
from everest.repositories.rdb.utils import set_metadata
from everest.repositories.utils import get_engine
from everest.repositories.utils import is_engine_initialized
from everest.repositories.utils import reset_engines
from everest.repositories.utils import set_engine
from everest.testing import Pep8CompliantTestCase
from everest.tests.complete_app.entities import MyEntity
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy.engine import create_engine
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.sql.expression import cast
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['RdbTestCase',
           ]


class RdbTestCase(Pep8CompliantTestCase):
    def test_rdb_engine_manager(self):
        key = 'test'
        self.assert_false(is_engine_initialized(key))
        eng = create_engine('sqlite://')
        set_engine(key, eng)
        self.assert_raises(ValueError, set_engine, key, eng)
        self.assert_true(is_engine_initialized(key))
        self.assert_true(get_engine(key) is eng)
        reset_engines()
        self.assert_false(is_engine_initialized(key))

    def test_metadata_manager(self):
        key = 'test'
        self.assert_false(is_metadata_initialized(key))
        md = MetaData()
        set_metadata(key, md)
        self.assert_raises(ValueError, set_metadata, key, md)
        self.assert_true(is_metadata_initialized(key))
        self.assert_true(get_metadata(key) is md)
        reset_metadata()
        self.assert_false(is_metadata_initialized(key))

    def test_mapper(self):
        class MyDerivedEntity(MyEntity):
            pass

        @implementer(IEntity)
        class MyEntityWithCustomId(object):

            def __init__(self, id=None): # redefining id pylint: disable=W0622
                self.__id = id
                self.__slug = str(id)

            @property
            def id(self):
                return str(self.__id)

            @property
            def slug(self):
                return str(self.__id)

        class MyPolymorphicEntity1(MyEntityWithCustomId):
            pass

        class MyPolymorphicEntity2(MyEntityWithCustomId):
            pass

        t1 = self._make_table(True)
        with self.assert_raises(ValueError) as cm:
            mpr = mapper(MyDerivedEntity, t1, id_attribute='my_id')
        msg_str = 'Attempting to overwrite the mapped'
        self.assert_true(str(cm.exception).startswith(msg_str))
        t2 = self._make_table(False)
        with self.assert_raises(ValueError) as cm:
            mpr = mapper(MyEntityWithCustomId, t2, id_attribute='my_id')
        msg_str = 'Attempting to overwrite the custom data'
        self.assert_true(str(cm.exception).startswith(msg_str))
        #
        slug_expr = lambda cls: as_slug_expression(cast(cls.id, String))
        mpr = mapper(MyDerivedEntity, t2,
                     id_attribute='my_id', slug_expression=slug_expr)
        self.assert_true(MyDerivedEntity.__dict__['slug'].expr
                         is slug_expr)
        self.assert_true(isinstance(MyDerivedEntity.slug, FunctionElement))
        mpr.dispose()
        # Test mapping polymorphic class with custom slug in the base class.
        base_mpr = mapper(MyEntityWithCustomId, t2,
                          polymorphic_on='my_type',
                          polymorphic_identity='base')
        poly_mpr1 = mapper(MyPolymorphicEntity1, inherits=base_mpr,
                           polymorphic_identity='derived')
        self.assert_true(isinstance(MyPolymorphicEntity1.__dict__['slug'],
                                    hybrid_descriptor))
        # We should not override the slug expression if we are inheriting
        # a hybrid descriptor from the base class.
        self.assert_raises(ValueError,
                           mapper, MyPolymorphicEntity2, inherits=base_mpr,
                           polymorphic_identity='derived',
                           slug_expression=slug_expr)
        base_mpr.dispose()
        poly_mpr1.dispose()

    def _make_table(self, with_id):
        md = MetaData()
        cols = [Column('my_id', Integer, primary_key=True),
                Column('my_type', String)]
        if with_id:
            cols.append(Column('id', Integer))
        return Table('my_table_with_id_col', md, *cols)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from datetime import datetime
from datetime import timedelta

from pyramid.compat import iteritems_
import pytest

from everest.querying.operators import UnaryOperator
from everest.querying.specifications import ConjunctionFilterSpecification
from everest.querying.specifications import DisjunctionFilterSpecification
from everest.querying.specifications import FilterSpecification
from everest.querying.specifications import FilterSpecificationGenerator
from everest.querying.specifications import NegationFilterSpecification
from everest.querying.specifications import asc
from everest.querying.specifications import cntd
from everest.querying.specifications import cnts
from everest.querying.specifications import desc
from everest.querying.specifications import ends
from everest.querying.specifications import eq
from everest.querying.specifications import ge
from everest.querying.specifications import gt
from everest.querying.specifications import le
from everest.querying.specifications import lt
from everest.querying.specifications import rng
from everest.querying.specifications import starts


__docformat__ = 'reStructuredText en'
__all__ = ['TestFilterSpecification',
           ]


TEXT_VALUE = 'Beta-2'
GREATER_THAN_TEXT_VALUE = 'Gamma-3'
LESS_THAN_TEXT_VALUE = 'Alpha-1'
TEXT_VALUE_LIST = [LESS_THAN_TEXT_VALUE, TEXT_VALUE,
                   GREATER_THAN_TEXT_VALUE]
NUMBER_VALUE = 40
GREATER_THAN_NUMBER_VALUE = NUMBER_VALUE + 1
LESS_THAN_NUMBER_VALUE = NUMBER_VALUE - 1
DATE_VALUE = datetime(1970, 1, 1)
GREATER_THAN_DATE_VALUE = DATE_VALUE + timedelta(1)
LESS_THAN_DATE_VALUE = DATE_VALUE - timedelta(1)
LIST_VALUES = [1, 2, 3, 4, 5]


class AlwaysTrueOperator(UnaryOperator):
    name = 'true'
    literal = 'T'

    @staticmethod
    def apply(arg): # pylint: disable=W0613
        return True


class AlwaysTrueFilterSpecification(FilterSpecification):

    operator = AlwaysTrueOperator

    def __init__(self):
        FilterSpecification.__init__(self)

    def is_satisfied_by(self, candidate):
        return self.operator.apply(candidate)

    def accept(self, visitor):
        pass


class AlwaysFalseOperator(UnaryOperator):
    name = 'false'
    literal = 'F'

    @staticmethod
    def apply(arg): # pylint: disable=W0613
        return False


class AlwaysFalseFilterSpecification(FilterSpecification):
    operator = AlwaysFalseOperator
    def is_satisfied_by(self, candidate):
        return False

    def accept(self, visitor):
        pass


class SpecificationCandidate(object):
    def __str__(self):
        attrs = ['%s: %s' % (k, getattr(self, k))
                 for k in self.__dict__
                 if not k.startswith('_')]
        return 'Candidate -> %s' % ', '.join(attrs)

    @classmethod
    def make_instance(self, **attributes):
        cand = SpecificationCandidate()
        for attr_name, attr_value in iteritems_(attributes):
            setattr(cand, attr_name, attr_value)
        return cand


@pytest.fixture
def specification_candidate_factory():
    return SpecificationCandidate.make_instance


@pytest.fixture
def specification_candidate(specification_candidate_factory): #pylint: disable=W0621
    return specification_candidate_factory(text_attr=TEXT_VALUE,
                                           number_attr=NUMBER_VALUE,
                                           date_attr=DATE_VALUE,
                                           list_attr=LIST_VALUES)


class TestFilterSpecification(object):

    def test_basics(self, filter_specification_factory):
        spec = filter_specification_factory.create_equal_to('foo', 'bar')
        assert spec == spec
        spec_other_attr = \
            filter_specification_factory.create_equal_to('bar', 'bar')
        assert spec != spec_other_attr
        spec_other_value = \
            filter_specification_factory.create_equal_to('foo', 'bar1')
        assert spec != spec_other_value
        str_str = '<%s op_name:' % spec.__class__.__name__
        assert str(spec)[:len(str_str)] == str_str

    @pytest.mark.parametrize('value,attr,outcome',
                             [(TEXT_VALUE, 'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE, 'text_attr', False),
                              (NUMBER_VALUE, 'number_attr', True),
                              (GREATER_THAN_NUMBER_VALUE, 'number_attr',
                               False),
                              (DATE_VALUE, 'date_attr', True),
                              (GREATER_THAN_DATE_VALUE, 'date_attr', False),
                              ])
    def test_equal_to(self, filter_specification_factory,
                      specification_candidate, # pylint: disable=W0621
                      attr, value, outcome):
        spec = filter_specification_factory.create_equal_to(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr,outcome',
                             [(LESS_THAN_TEXT_VALUE, 'text_attr', True),
                              (TEXT_VALUE, 'text_attr', False),
                              (LESS_THAN_NUMBER_VALUE, 'number_attr', True),
                              (NUMBER_VALUE, 'number_attr', False),
                              (LESS_THAN_DATE_VALUE, 'date_attr', True),
                              (DATE_VALUE, 'date_attr', False),
                              ])
    def test_greater_than(self, filter_specification_factory,
                          specification_candidate, # pylint: disable=W0621
                          attr, value, outcome):
        spec = filter_specification_factory.create_greater_than(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr,outcome',
                             [(GREATER_THAN_TEXT_VALUE, 'text_attr', True),
                              (TEXT_VALUE, 'text_attr', False),
                              (GREATER_THAN_NUMBER_VALUE, 'number_attr',
                               True),
                              (NUMBER_VALUE, 'number_attr', False),
                              (GREATER_THAN_DATE_VALUE, 'date_attr', True),
                              (DATE_VALUE, 'date_attr', False),
                              ])
    def test_less_than(self, filter_specification_factory,
                       specification_candidate, # pylint: disable=W0621
                       attr, value, outcome):
        spec = filter_specification_factory.create_less_than(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr,outcome',
                             [(LESS_THAN_TEXT_VALUE, 'text_attr', True),
                              (TEXT_VALUE, 'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE, 'text_attr', False),
                              (LESS_THAN_NUMBER_VALUE, 'number_attr', True),
                              (NUMBER_VALUE, 'number_attr', True),
                              (GREATER_THAN_NUMBER_VALUE, 'number_attr',
                               False),
                              (LESS_THAN_DATE_VALUE, 'date_attr', True),
                              (DATE_VALUE, 'date_attr', True),
                              (GREATER_THAN_DATE_VALUE, 'date_attr', False),
                              ])
    def test_greater_than_or_equal_to(self,
                                      filter_specification_factory,
                                      specification_candidate, # pylint: disable=W0621
                                      attr, value, outcome):
        spec = \
          filter_specification_factory.create_greater_than_or_equal_to(attr,
                                                                       value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr,outcome',
                             [(GREATER_THAN_TEXT_VALUE, 'text_attr', True),
                              (TEXT_VALUE, 'text_attr', True),
                              (LESS_THAN_TEXT_VALUE, 'text_attr', False),
                              (GREATER_THAN_NUMBER_VALUE, 'number_attr',
                               True),
                              (NUMBER_VALUE, 'number_attr', True),
                              (LESS_THAN_NUMBER_VALUE, 'number_attr', False),
                              (GREATER_THAN_DATE_VALUE, 'date_attr', True),
                              (DATE_VALUE, 'date_attr', True),
                              (LESS_THAN_DATE_VALUE, 'date_attr', False),
                              ])
    def test_less_than_or_equal_to(self, filter_specification_factory,
                                   specification_candidate, # pylint: disable=W0621
                                   attr, value, outcome):
        spec = \
            filter_specification_factory.create_less_than_or_equal_to(attr,
                                                                      value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    def test_in_range_basics(self, filter_specification_factory):
        spec = filter_specification_factory.create_in_range('foo',
                                                            ('bar0', 'bar1'))
        assert spec.from_value == 'bar0'
        assert spec.to_value == 'bar1'
        assert spec == spec
        spec_other_value = \
            filter_specification_factory.create_in_range('foo',
                                                         ('bar0', 'bar2'))
        assert spec != spec_other_value
        spec_other_attr = \
            filter_specification_factory.create_in_range('bar',
                                                         ('bar0', 'bar1'))
        assert spec != spec_other_attr

    @pytest.mark.parametrize('value1,value2,attr,outcome',
                             [(LESS_THAN_TEXT_VALUE, GREATER_THAN_TEXT_VALUE,
                               'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE, LESS_THAN_TEXT_VALUE,
                               'text_attr', False),
                              (LESS_THAN_NUMBER_VALUE,
                               GREATER_THAN_NUMBER_VALUE, 'number_attr',
                               True),
                              (GREATER_THAN_DATE_VALUE,
                               LESS_THAN_DATE_VALUE, 'date_attr', False),
                              (LESS_THAN_DATE_VALUE, GREATER_THAN_DATE_VALUE,
                               'date_attr', True),
                              (GREATER_THAN_DATE_VALUE, LESS_THAN_DATE_VALUE,
                               'date_attr', False),
                              ])
    def test_in_range(self, filter_specification_factory,
                      specification_candidate, # pylint: disable=W0621
                      attr, value1, value2, outcome):
        spec = filter_specification_factory.create_in_range(attr,
                                                            (value1, value2))
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr,outcome',
                             [(TEXT_VALUE[0], 'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE[0], 'text_attr',
                               False),
                              (LIST_VALUES[0], 'list_attr', True),
                              (LIST_VALUES[-1], 'list_attr', False),
                              ])
    def test_starts_with(self, filter_specification_factory,
                         specification_candidate, # pylint: disable=W0621
                         attr, value, outcome):
        spec = filter_specification_factory.create_starts_with(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr',
                             [(NUMBER_VALUE, 'number_attr'),
                              (DATE_VALUE, 'date_attr')
                              ])
    def test_starts_with_raises(self,
                                filter_specification_factory,
                                specification_candidate, # pylint: disable=W0621
                                attr, value):
        spec = filter_specification_factory.create_starts_with(attr, value)
        with pytest.raises(TypeError):
            spec.is_satisfied_by(specification_candidate)

    @pytest.mark.parametrize('value,attr,outcome',
                             [(TEXT_VALUE[-1], 'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE[-1], 'text_attr',
                               False),
                              (LIST_VALUES[-1], 'list_attr', True),
                              (LIST_VALUES[0], 'list_attr', False),
                              ])
    def test_ends_with(self, filter_specification_factory,
                       specification_candidate, # pylint: disable=W0621
                       attr, value, outcome):
        spec = filter_specification_factory.create_ends_with(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr',
                             [(NUMBER_VALUE, 'number_attr'),
                              (DATE_VALUE, 'date_attr')
                              ])
    def test_ends_with_raises(self,
                              filter_specification_factory,
                              specification_candidate, # pylint: disable=W0621
                              attr, value):
        spec = filter_specification_factory.create_ends_with(attr, value)
        with pytest.raises(TypeError):
            spec.is_satisfied_by(specification_candidate)

    @pytest.mark.parametrize('value,attr,outcome',
                             [(TEXT_VALUE[2], 'text_attr', True),
                              (GREATER_THAN_TEXT_VALUE[-1], 'text_attr',
                               False),
                              (LIST_VALUES[2], 'list_attr', True),
                              (-1, 'list_attr', False),
                              ])
    def test_contains(self, filter_specification_factory,
                      specification_candidate, # pylint: disable=W0621
                      attr, value, outcome):
        spec = filter_specification_factory.create_contains(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr',
                             [(NUMBER_VALUE, 'number_attr'),
                              (DATE_VALUE, 'date_attr'),
                              ])
    def test_contains_raises(self,
                              filter_specification_factory,
                              specification_candidate, # pylint: disable=W0621
                              attr, value):
        spec = filter_specification_factory.create_contains(attr, value)
        with pytest.raises(TypeError):
            spec.is_satisfied_by(specification_candidate)

    @pytest.mark.parametrize('value,attr,outcome',
                             [(TEXT_VALUE_LIST, 'text_attr', True),
                              (LIST_VALUES, 'text_attr', False),
                              ])
    def test_contained(self, filter_specification_factory,
                      specification_candidate, # pylint: disable=W0621
                      attr, value, outcome):
        spec = filter_specification_factory.create_contained(attr, value)
        assert spec.is_satisfied_by(specification_candidate) is outcome

    @pytest.mark.parametrize('value,attr',
                             [(NUMBER_VALUE, 'number_attr'),
                              (DATE_VALUE, 'date_attr'),
                              ])
    def test_contained_raises(self,
                              filter_specification_factory,
                              specification_candidate, # pylint: disable=W0621
                              attr, value):
        spec = filter_specification_factory.create_contained(attr, value)
        with pytest.raises(TypeError):
            spec.is_satisfied_by(specification_candidate)

    def test_conjunction_basics(self, filter_specification_factory):
        always_true_spec = AlwaysTrueFilterSpecification()
        always_false_spec = AlwaysFalseFilterSpecification()
        spec = \
            filter_specification_factory.create_conjunction(always_true_spec,
                                                            always_false_spec)
        assert spec == spec
        other_spec = \
            filter_specification_factory.create_conjunction(always_false_spec,
                                                            always_true_spec)
        assert spec != other_spec
        str_str = '<%s left_spec:' % spec.__class__.__name__
        assert str(spec)[:len(str_str)] == str_str

    @pytest.mark.parametrize('left_spec,right_spec,outcome',
                             [(AlwaysTrueFilterSpecification(),
                               AlwaysTrueFilterSpecification(), True),
                              (AlwaysFalseFilterSpecification(),
                               AlwaysTrueFilterSpecification(), False),
                              (AlwaysTrueFilterSpecification(),
                               AlwaysFalseFilterSpecification(), False),
                              (AlwaysFalseFilterSpecification(),
                               AlwaysFalseFilterSpecification(), False),
                              ])
    def test_conjunction(self, filter_specification_factory,
                         specification_candidate_factory, #pylint: disable=W0621
                         left_spec, right_spec, outcome):
        spec = filter_specification_factory.create_conjunction(left_spec,
                                                               right_spec)
        cand = specification_candidate_factory()
        assert spec.is_satisfied_by(cand) is outcome

    @pytest.mark.parametrize('left_spec,right_spec,outcome',
                             [(AlwaysFalseFilterSpecification(),
                               AlwaysTrueFilterSpecification(), True),
                              (AlwaysTrueFilterSpecification(),
                               AlwaysFalseFilterSpecification(), True),
                              (AlwaysTrueFilterSpecification(),
                               AlwaysTrueFilterSpecification(), True),
                              (AlwaysFalseFilterSpecification(),
                               AlwaysFalseFilterSpecification(), False),
                              ])
    def test_disjunction(self, filter_specification_factory,
                         specification_candidate_factory, #pylint: disable=W0621
                         left_spec, right_spec, outcome):
        spec = filter_specification_factory.create_disjunction(left_spec,
                                                               right_spec)
        cand = specification_candidate_factory()
        assert spec.is_satisfied_by(cand) is outcome

    def test_negation_basics(self, filter_specification_factory):
        af_spec = AlwaysFalseFilterSpecification()
        spec = filter_specification_factory.create_negation(af_spec)
        assert spec == spec
        at_spec = AlwaysTrueFilterSpecification()
        other_spec = filter_specification_factory.create_negation(at_spec)
        assert spec != other_spec
        str_str = '<%s wrapped_spec:' % spec.__class__.__name__
        assert str(spec)[:len(str_str)] == str_str

    @pytest.mark.parametrize('wrapped_spec,outcome',
                             [(AlwaysFalseFilterSpecification(), True),
                              (AlwaysTrueFilterSpecification(), False)
                              ])
    def test_negation(self, filter_specification_factory,
                      specification_candidate_factory, #pylint: disable=W0621
                      wrapped_spec, outcome):
        spec = filter_specification_factory.create_negation(wrapped_spec)
        cand = specification_candidate_factory()
        assert spec.is_satisfied_by(cand) is outcome

    def test_order_basics(self, order_specification_factory):
        spec = order_specification_factory.create_ascending('foo')
        str_str = '<%s attr_name:' % spec.__class__.__name__
        assert str(spec)[:len(str_str)] == str_str

    @pytest.mark.parametrize('attr',
                             ['number_attr', 'text_attr'])
    def test_order_ascending(self, order_specification_factory,
                             specification_candidate_factory, #pylint: disable=W0621
                             attr):
        spec = order_specification_factory.create_ascending(attr)
        first_candidate = \
            specification_candidate_factory(number_attr=0, text_attr='a')
        second_candidate = \
            specification_candidate_factory(number_attr=1, text_attr='b')
        assert spec.attr_name == attr
        assert not spec.eq(first_candidate, second_candidate)
        assert spec.ne(first_candidate, second_candidate)
        assert spec.lt(first_candidate, second_candidate)
        assert not spec.ge(first_candidate, second_candidate)
        assert spec.le(first_candidate, second_candidate)
        assert not spec.gt(first_candidate, second_candidate)

    @pytest.mark.parametrize('attr',
                             ['number_attr', 'text_attr'])
    def test_order_descending(self, order_specification_factory,
                              specification_candidate_factory, #pylint: disable=W0621
                              attr):
        spec = order_specification_factory.create_descending(attr)
        first_candidate = \
            specification_candidate_factory(number_attr=0, text_attr='a')
        second_candidate = \
            specification_candidate_factory(number_attr=1, text_attr='b')
        assert spec.attr_name == attr
        assert not spec.eq(first_candidate, second_candidate)
        assert spec.ne(first_candidate, second_candidate)
        assert not spec.lt(first_candidate, second_candidate)
        assert spec.ge(first_candidate, second_candidate)
        assert not spec.le(first_candidate, second_candidate)
        assert spec.gt(first_candidate, second_candidate)

    def test_order_natural(self, order_specification_factory,
                           specification_candidate_factory): #pylint: disable=W0621
        text_spec = order_specification_factory.create_natural('text_attr')
        first_candidate = \
            specification_candidate_factory(number_attr=0, text_attr='a10')
        second_candidate = \
            specification_candidate_factory(number_attr=1, text_attr='a9')
        assert not text_spec.lt(first_candidate, second_candidate)
        number_spec = \
                order_specification_factory.create_natural('number_attr')
        assert number_spec.lt(first_candidate, second_candidate)

    def test_order_conjunction(self, order_specification_factory,
                               specification_candidate_factory): #pylint: disable=W0621
        text_spec = order_specification_factory.create_natural('text_attr')
        number_spec = \
                order_specification_factory.create_natural('number_attr')
        conj_spec = \
            order_specification_factory.create_conjunction(number_spec,
                                                           text_spec)
        first_candidate = \
            specification_candidate_factory(number_attr=0, text_attr='a')
        second_candidate = \
            specification_candidate_factory(number_attr=0, text_attr='b')
        str_str = '<%s left_spec:' % conj_spec.__class__.__name__
        assert str(conj_spec)[:len(str_str)] == str_str
        assert conj_spec.lt(first_candidate, second_candidate)
        assert conj_spec.le(first_candidate, second_candidate)
        assert not conj_spec.eq(first_candidate, second_candidate)
        assert conj_spec.cmp(first_candidate, second_candidate) == -1
        inv_conj_spec = \
            order_specification_factory.create_conjunction(text_spec,
                                                           number_spec)
        assert inv_conj_spec.lt(first_candidate, second_candidate)
        assert inv_conj_spec.le(first_candidate, second_candidate)
        assert not inv_conj_spec.eq(first_candidate, second_candidate)
        assert inv_conj_spec.cmp(first_candidate, second_candidate) == -1


class TestSpecificationGenerator(object):

    @pytest.mark.parametrize('attrs,generator',
                             [(dict(number_attr=NUMBER_VALUE,
                                    text_attr=TEXT_VALUE), eq),
                              (dict(text_attr=TEXT_VALUE[0]), starts),
                              (dict(text_attr=TEXT_VALUE[-1]), ends),
                              (dict(number_attr=NUMBER_VALUE + 1), lt),
                              (dict(number_attr=NUMBER_VALUE), le),
                              (dict(number_attr=NUMBER_VALUE - 1), gt),
                              (dict(number_attr=NUMBER_VALUE), ge),
                              (dict(text_attr=TEXT_VALUE[1:2]), cnts),
                              (dict(text_attr=TEXT_VALUE), cntd),
                              (dict(number_attr=(NUMBER_VALUE - 1,
                                                 NUMBER_VALUE + 1)), rng),
                               ])
    def test_plain_generators(self, configurator,
                              specification_candidate, attrs, #pylint: disable=W0621
                              generator):
        configurator.begin()
        try:
            spec = generator(**attrs)
            if len(attrs) > 1:
                assert isinstance(spec, ConjunctionFilterSpecification)
            assert spec.is_satisfied_by(specification_candidate)
        finally:
            configurator.end()

    @pytest.mark.parametrize('attrs,generator,outcome',
                             [(('number_attr', 'text_attr'), asc, True),
                              (('number_attr', 'text_attr'), desc, False),
                              ])
    def test_order_generators(self, configurator,
                              specification_candidate_factory, #pylint: disable=W0621
                              attrs, generator, outcome):
        first_candidate = \
             specification_candidate_factory(number_attr=NUMBER_VALUE,
                                             text_attr=TEXT_VALUE)
        second_candidate = \
             specification_candidate_factory(number_attr=NUMBER_VALUE,
                                             text_attr=GREATER_THAN_TEXT_VALUE)
        configurator.begin()
        try:
            spec = generator(*attrs)
            assert spec.lt(first_candidate, second_candidate) is outcome
        finally:
            configurator.end()

    def test_instantiating_generator(self, filter_specification_factory,
                                     specification_candidate): #pylint: disable=W0621
        gen = FilterSpecificationGenerator(filter_specification_factory)
        spec = gen.lt(number_attr=NUMBER_VALUE + 1) \
                & gen.gt(number_attr=NUMBER_VALUE - 1)
        assert spec.is_satisfied_by(specification_candidate)

    def test_generator_or(self, specification_candidate): #pylint: disable=W0621
        spec = lt(number_attr=NUMBER_VALUE + 1) \
                | gt(number_attr=NUMBER_VALUE + 1)
        assert spec.is_satisfied_by(specification_candidate)

    def test_and(self, specification_candidate): #pylint: disable=W0621
        spec = eq(number_attr=NUMBER_VALUE) & eq(text_attr=TEXT_VALUE)
        assert isinstance(spec, ConjunctionFilterSpecification)
        assert spec.is_satisfied_by(specification_candidate)

    def test_or(self, specification_candidate): #pylint: disable=W0621
        spec = eq(number_attr=NUMBER_VALUE - 1) | eq(text_attr=TEXT_VALUE)
        assert isinstance(spec, DisjunctionFilterSpecification)
        assert spec.is_satisfied_by(specification_candidate)

    def test_not(self, specification_candidate): #pylint: disable=W0621
        spec = ~eq(number_attr=NUMBER_VALUE - 1)
        assert isinstance(spec, NegationFilterSpecification)
        assert spec.is_satisfied_by(specification_candidate)



"""
Testing utilities for the rdb backend.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 26, 2013.
"""
from functools import update_wrapper

from pyramid.compat import iteritems_
from pyramid.threadlocal import get_current_registry

from everest.repositories.constants import REPOSITORY_TYPES
from everest.repositories.interfaces import IRepositoryManager
from everest.repositories.rdb.session import ScopedSessionMaker as Session
from everest.repositories.rdb.utils import reset_metadata
from everest.repositories.utils import get_engine


__docformat__ = 'reStructuredText en'
__all__ = ['RdbContextManager',
           'RdbTestCaseMixin',
           'check_attributes',
           'persist',
           'with_rdb',
           ]


def check_attributes(test_object, attribute_map):
    """
    Utility function to test whether the test object attributes match the
    expected ones (given the dictionary).

    :param test_object: a test object
    :param attribute_map: a dictionary with key = attribute name
            and value = expected value for this attribute
    """
    for attr_name, exp_val in iteritems_(attribute_map):
        obj_val = getattr(test_object, attr_name)
        if obj_val != exp_val:
            raise AssertionError('Values for attribute %s differ!'
                                 % attr_name)


def persist(session, entity_class, attribute_map,
            do_attribute_check=True):
    """
    Utility function which creates an object of the given class with the
    given attribute map, commits it to the backend, reloads it from the
    backend and then tests if the attributes compare equal.

    :param entity_class: class inheriting from
      :class:`everest.entities.base.Entity`
    :param attribute_map: a dictionary containint attribute names as keys
      and expected attribute values as values. The attribute map must
      contain all mandatory attributes required for instantiation.
    """
    # Instantiate.
    entity = entity_class(**attribute_map)
    session.add(entity_class, entity)
    session.commit()
    session.refresh(entity)
    entity_id = entity.id
    # Assure a new object is loaded to test if persisting worked.
    session.expunge(entity)
    del entity
    query = session.query(entity_class)
    fetched_entity = query.filter_by(id=entity_id).one()
    if do_attribute_check:
        check_attributes(fetched_entity, attribute_map)


class RdbContextManager(object):
    """
    Context manager for RDB tests.

    Configures the entity repository to use the RDB implementation as
    a default, sets up an outer transaction before the test is run and rolls
    this transaction back after the test has finished.
    """
    def __init__(self, autoflush=True, engine_name=None):
        if engine_name is None:
            # Use the name of the default RDB repository for engine lookup.
            engine_name = REPOSITORY_TYPES.RDB
        self.__autoflush = autoflush
        self.__engine_name = engine_name
        self.__connection = None
        self.__transaction = None
        self.__session = None
        self.__repo = None
        self.__old_autoflush_flag = None
        self.__old_join_transaction_flag = None

    def __enter__(self):
        # We set up an outer transaction that allows us to roll back all
        # changes (including commits) the unittest may want to make.
        engine = get_engine(self.__engine_name)
        self.__connection = engine.connect()
        self.__transaction = self.__connection.begin()
        reg = get_current_registry()
        repo_mgr = reg.getUtility(IRepositoryManager)
        self.__repo = repo_mgr.get(self.__engine_name)
        # Configure the autoflush behavior of the session.
        if Session.registry.has():
            self.__old_autoflush_flag = Session.autoflush # pylint:disable=E1101
        else:
            self.__old_autoflush_flag = self.__autoflush
        Session.remove()
        Session.configure(autoflush=self.__autoflush)
        # Create a new session for the tests.
        Session.configure(bind=self.__connection)
        self.__old_join_transaction_flag = self.__repo.join_transaction
        self.__repo.join_transaction = False
        self.__session = self.__repo.session_factory()
        return self.__session

    def __exit__(self, ext_type, value, tb):
        # Roll back the outer transaction and close the connection.
        self.__session.close()
        self.__transaction.rollback()
        self.__connection.close()
        # Remove the session we created.
        Session.remove()
        # Restore flags.
        Session.configure(autoflush=self.__old_autoflush_flag,
                          bind=get_engine(self.__engine_name))
        self.__repo.join_transaction = self.__old_join_transaction_flag


def with_rdb(autoflush=True, init_callback=None):
    """
    Decorator for tests which uses a :class:`RdbContextManager` for the
    call to the decorated test function.
    """
    def decorate(func):
        def wrap(*args, **kw):
            with RdbContextManager(autoflush=autoflush,
                                   init_callback=init_callback):
                func(*args, **kw)
        return update_wrapper(wrap, func)
    return decorate


class RdbTestCaseMixin(object):
    def tear_down(self):
        super(RdbTestCaseMixin, self).tear_down()
        Session.remove()

#    @classmethod
#    def setup_class(cls):
#        base_cls = super(RdbTestCaseMixin, cls)
#        try:
#            base_cls.setup_class()
#        except AttributeError:
#            pass
#        Session.remove()
#        assert not Session.registry.has()
#        reset_metadata()
#        reset_engines()

    @classmethod
    def teardown_class(cls):
        base_cls = super(RdbTestCaseMixin, cls)
        try:
            base_cls.teardown_class()
        except AttributeError:
            pass
        Session.remove()
        assert not Session.registry.has()
        reset_metadata()


def no_autoflush(scoped_session=None):
    """
    Decorator to disable autoflush on the session for the duration of a
    test call. Uses the scoped session from the :mod:`everest.db` module
    as default.

    Adapted from
    http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DisableAutoflush
    """
    if scoped_session is None:
        scoped_session = Session
    def decorate(fn):
        def wrap(*args, **kw):
            session = scoped_session()
            autoflush = session.autoflush
            session.autoflush = False
            try:
                return fn(*args, **kw)
            finally:
                session.autoflush = autoflush
        return update_wrapper(wrap, fn)
    return decorate



"""
Custom exceptions.

This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 13, 2010.
"""
__docformat__ = 'reStructuredText en'
__all__ = ['MultipleResultsException',
           'NoResultsException',
           'UnsupportedOperationException',
           ]


class UnsupportedOperationException(Exception):
    """
    Raise this to indicate that the requested operation is not supported.
    """


class NoResultsException(Exception):
    """
    Raised when no result was found when at least one was expected.
    """


class MultipleResultsException(Exception):
    """
    Raised when more than one item was found where at most one was expected. 
    """




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 7, 2012.
"""
from collections import OrderedDict
import datetime

import pytest
from pytz import timezone

from everest.constants import RESOURCE_KINDS
from everest.representers.attributes import MappedAttribute
from everest.representers.csv import CsvLinkedDataElement
from everest.representers.dataelements import DataElementAttributeProxy
from everest.representers.dataelements import SimpleCollectionDataElement
from everest.representers.dataelements import SimpleLinkedDataElement
from everest.representers.dataelements import SimpleMemberDataElement
from everest.resources.attributes import get_resource_class_attribute
from everest.tests.complete_app.resources import MyEntityMember

from pyramid.compat import iteritems_


__docformat__ = 'reStructuredText en'
__all__ = ['TestDataElements',
           ]


class TestDataElements(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_member_data_element(self):
        data_el = SimpleMemberDataElement.create()
        # Test nesteds.
        assert list(data_el.nesteds.keys()) == []
        parent_data_el = SimpleMemberDataElement.create()
        rc_nested_attr = \
            get_resource_class_attribute(MyEntityMember, 'parent')
        mp_nested_attr = MappedAttribute(rc_nested_attr)
        data_el.set_nested(mp_nested_attr, parent_data_el)
        assert data_el.get_nested(mp_nested_attr) is parent_data_el
        assert list(data_el.nesteds.keys()) == ['parent']
        # Test terminals.
        assert list(data_el.terminals.keys()) == []
        utc = timezone('UTC')
        ldt = datetime.datetime(2012, 8, 29, 16, 20, 0, tzinfo=utc)
        term_attr_data = OrderedDict(text='foo',
                                     number=0,
                                     date_time=ldt)
        for term_attr_name, term_attr_value in term_attr_data.items():
            rc_attr = get_resource_class_attribute(MyEntityMember,
                                                   term_attr_name)
            mp_attr = MappedAttribute(rc_attr)
            # Check setting to None value.
            data_el.set_terminal(mp_attr, None)
            assert data_el.get_terminal(mp_attr) is None
            data_el.set_terminal_converted(mp_attr, None)
            assert data_el.get_terminal(mp_attr) is None
            # Check setting to value.
            data_el.set_terminal(mp_attr, term_attr_value)
            assert data_el.get_terminal(mp_attr) == term_attr_value
            rpr_val = data_el.get_terminal_converted(mp_attr)
            data_el.set_terminal_converted(mp_attr, rpr_val)
            assert data_el.get_terminal(mp_attr) == term_attr_value
        assert list(data_el.terminals.keys()) == list(term_attr_data.keys())
        # Printing.
        prt_str = str(data_el)
        assert prt_str.startswith(data_el.__class__.__name__)
        assert prt_str.endswith(')')
        # Attribute iteration.
        for attr_name, attr_value in iteritems_(term_attr_data):
            assert data_el.get_attribute(attr_name) == attr_value
        for de_attr_name, de_attr_value in data_el.iterator():
            if de_attr_name in term_attr_data:
                assert de_attr_value == term_attr_data[de_attr_name]

    def test_printing_with_none_value(self):
        data_el = SimpleMemberDataElement.create()
        assert list(data_el.terminals.keys()) == []
        rc_attr = get_resource_class_attribute(MyEntityMember, 'text')
        mp_attr = MappedAttribute(rc_attr)
        data_el.set_terminal(mp_attr, None) # Need one None attr value.
        prt_str = str(data_el)
        assert prt_str.startswith(data_el.__class__.__name__)
        assert prt_str.endswith(')')

    def test_collection_data_element(self):
        coll_data_el = SimpleCollectionDataElement.create()
        mb_data_el1 = SimpleMemberDataElement.create()
        mb_data_el2 = SimpleMemberDataElement.create()
        coll_data_el.add_member(mb_data_el1)
        coll_data_el.add_member(mb_data_el2)
        assert len(coll_data_el) == 2
        assert coll_data_el.members == [mb_data_el1, mb_data_el2]
        # Printing.
        prt_str = str(coll_data_el)
        assert prt_str.startswith(coll_data_el.__class__.__name__)
        assert prt_str.endswith(']')

    def test_linked_collection_data_element(self, collection):
        agg = collection.get_aggregate()
        with pytest.raises(ValueError):
            CsvLinkedDataElement.create_from_resource(agg)
        data_el = CsvLinkedDataElement.create_from_resource(collection)
        assert data_el.get_title() == 'Collection of MyEntityMember'
        assert data_el.get_url().find('/my-entities/') != -1
        assert data_el.get_kind() == RESOURCE_KINDS.COLLECTION
        assert data_el.get_relation() == 'http://test.org/myentity-collection'
        # Can not use data element attribute proxy with a link.
        with pytest.raises(ValueError):
            DataElementAttributeProxy(data_el)
        assert str(data_el).startswith(data_el.__class__.__name__)

    def test_linked_member_data_element(self):
        mb_data_el = SimpleLinkedDataElement.create('http://dummy',
                                                    RESOURCE_KINDS.MEMBER,
                                                    id=0)
        assert mb_data_el.get_kind() == RESOURCE_KINDS.MEMBER
        assert mb_data_el.get_id() == 0



"""
Get collection view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 7, 2011.
"""
from copy import deepcopy

from everest.batch import Batch
from everest.resources.base import Link
from everest.url import UrlPartsConverter
from everest.utils import get_traceback
from everest.views.base import GetResourceView

__docformat__ = "reStructuredText en"
__all__ = ['GetCollectionView',
           ]


class GetCollectionView(GetResourceView):
    """
    View for GET requests on collection resources.

    If the request is sucessful, the server responds with status HTTP OK.
    """
    def _prepare_resource(self):
        try:
            self.__filter_collection()
            self.__order_collection()
            self.__slice_collection()
        except ValueError as err:
            result = self._handle_unknown_exception(err.args[0],
                                                    get_traceback())
        else:
            needs_default_order = self.context.order is None
            if needs_default_order:
                # Make sure we have defined an ordering on the collection
                # to guarantee an order on the result set. This should not
                # be reflected in the links' URLs.
                self.context.order = deepcopy(self.context.default_order)
            # Build batch links.
            batch = self.__create_batch()
            self_link = Link(self.context, 'self', self.context.title)
            self.context.add_link(self_link)
            if batch.index > 0:
                first_link = self.__create_nav_link(batch.first, 'first',
                                                    not needs_default_order)
                self.context.add_link(first_link)
            if not batch.previous is None:
                prev_link = self.__create_nav_link(batch.previous, 'previous',
                                                   not needs_default_order)
                self.context.add_link(prev_link)
            if not batch.next is None:
                next_link = self.__create_nav_link(batch.next, 'next',
                                                   not needs_default_order)
                self.context.add_link(next_link)
            if not batch.index == batch.number - 1:
                last_link = self.__create_nav_link(batch.last, 'last',
                                                   not needs_default_order)
                self.context.add_link(last_link)
            result = self.context
        return result

    def __create_batch(self):
        start = self.context.slice.start
        size = self.context.slice.stop - start
        total_size = len(self.context)
        return Batch(start, size, total_size)

    def __create_nav_link(self, batch, rel, reset_order):
        coll_clone = self.context.clone()
        if reset_order:
            coll_clone.order = None
        coll_clone.slice = slice(batch.start,
                                 batch.start + batch.size)
        return Link(coll_clone, rel, self.context.title)

    def __filter_collection(self):
        query_string = self.request.params.get('q')
        if not query_string is None:
            filter_spec = \
                UrlPartsConverter.make_filter_specification(query_string)
            self.context.filter = filter_spec

    def __order_collection(self):
        order_string = self.request.params.get('sort')
        if not order_string is None:
            order_spec = \
                UrlPartsConverter.make_order_specification(order_string)
            self.context.order = order_spec

    def __slice_collection(self):
        start_string = self.request.params.get('start')
        if start_string is None:
            start_string = '0'
        size_string = self.request.params.get('size')
        if size_string is None:
            size_string = str(self.context.default_limit)
        slice_key = UrlPartsConverter.make_slice_key(start_string, size_string)
        if not self.context.max_limit is None \
           and slice_key.stop - slice_key.start > self.context.max_limit:
            # Apply maximum batch size, if necessary.
            slice_key = slice(slice_key.start,
                              slice_key.start + self.context.max_limit)
        self.context.slice = slice_key



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 21, 2011.
"""

from everest.testing import EverestIni
from everest.testing import Pep8CompliantTestCase
from shutil import rmtree
from tempfile import mkdtemp
from tempfile import mktemp
from everest.compat import open_text

__docformat__ = 'reStructuredText en'
__all__ = []


INI = """\
[DEFAULT]
db_server = my_db_server
db_port = 5432
db_user = my_db_user
db_password = pwd123
db_name = my_db_name

[app:mysimple_app]
db_string = postgresql://%(db_user)s:%(db_password)s@%(db_server)s:%(db_port)s/%(db_name)s
db_echo = false
"""

class TestingTestCase(Pep8CompliantTestCase):
    ini_section_name = 'app:mysimple_app'

    def set_up(self):
        self.__testdir = mkdtemp()
        fn = mktemp(suffix="ini", dir=self.__testdir)
        ini_file = open_text(fn)
        ini_file.write(INI)
        ini_file.close()
        self.ini_file_path = fn

    def tear_down(self):
        rmtree(self.__testdir)

    def test_ini_file_read(self):
        ini = EverestIni(self.ini_file_path)
        ini_marker = self.ini_section_name
        db_string = ini.get_setting(ini_marker, 'db_string')
        self.assert_equal(
                db_string,
                'postgresql://my_db_user:pwd123@my_db_server:5432/my_db_name')



"""
Querying utilities.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 19, 2011.
"""
from everest.querying.interfaces import IFilterSpecificationFactory
from everest.querying.interfaces import IOrderSpecificationFactory
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['get_filter_specification_factory',
           'get_order_specification_factory',
           ]


def get_filter_specification_factory():
    """
    Returns the object registered as filter specification factory utility.

    :returns: object implementing
        :class:`everest.querying.interfaces.IFilterSpecificationFactory`
    """
    reg = get_current_registry()
    return reg.getUtility(IFilterSpecificationFactory)


def get_order_specification_factory():
    """
    Returns the object registered as order specification factory utility.

    :returns: object implementing
        :class:`everest.querying.interfaces.IOrderSpecificationFactory`
    """
    reg = get_current_registry()
    return reg.getUtility(IOrderSpecificationFactory)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 2, 2012.
"""
from collections import OrderedDict
import os

from pyramid.compat import binary_type
from pyramid.compat import bytes_
from pyramid.compat import text_type
import pytest

from everest.constants import RESOURCE_KINDS
from everest.entities.attributes import get_domain_class_attribute
from everest.mime import AtomMime
from everest.mime import CsvMime
from everest.mime import JsonMime
from everest.mime import XmlMime
from everest.querying.utils import get_filter_specification_factory
from everest.querying.utils import get_order_specification_factory
from everest.representers.attributes import MappedAttribute
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import REPR_NAME_OPTION
from everest.representers.config import RepresenterConfiguration
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.representers.csv import CsvData
from everest.representers.csv import CsvResourceRepresenter
from everest.representers.dataelements import CollectionDataElement
from everest.representers.dataelements import MemberDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.interfaces import IMemberDataElement
from everest.representers.interfaces import IRepresenterRegistry
from everest.representers.json import JsonDataTreeTraverser
from everest.representers.traversal import \
                        DataElementBuilderRepresentationDataVisitor
from everest.representers.utils import as_representer
from everest.representers.xml import NAMESPACE_MAPPING_OPTION
from everest.representers.xml import XML_NAMESPACE_OPTION
from everest.representers.xml import XML_PREFIX_OPTION
from everest.representers.xml import XML_SCHEMA_OPTION
from everest.representers.xml import XML_TAG_OPTION
from everest.resources.attributes import get_resource_class_attribute
from everest.resources.link import Link
from everest.resources.staging import create_staging_collection
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.resources import MyEntityMember
from everest.tests.complete_app.resources import MyEntityParentMember
from zope.interface import Interface # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['TestAtomRepresenter',
           'TestAttributes',
           'TestCsvData',
           'TestCsvRepresenter',
           'TestJsonRepresenter',
           'TestRepresenterConfiguration',
           'TestRepresenterConfigurationNoTypes',
           'TestRepresenterConfigurationOldStyleAttrs',
           'TestRepresenterRegistry',
           'TestUpdateResourceFromData',
           'TestXmlRepresenter',
           ]


class TestRepresenterRegistry(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_register_representer_class(self, configurator):
        configurator.begin()
        try:
            rpr_reg = configurator.get_registered_utility(IRepresenterRegistry)
            with pytest.raises(ValueError):
                rpr_reg.register_representer_class(CsvResourceRepresenter)
        finally:
            configurator.end()

    def test_register_representer(self, configurator):
        class MyMime(object):
            mime_string = 'application/mymime'
            file_extension = '.mymime'
        configurator.begin()
        try:
            rpr_reg = configurator.get_registered_utility(IRepresenterRegistry)
            with pytest.raises(ValueError) as cm:
                rpr_reg.register(MyEntity, CsvMime)
            exc_msg = 'Representers can only be registered for resource classes'
            assert str(cm.value).startswith(exc_msg)
            with pytest.raises(ValueError) as cm:
                rpr_reg.register(MyEntityMember, MyMime)
            exc_msg = 'No representer class has been registered for content type'
            assert str(cm.value).startswith(exc_msg)
        finally:
            configurator.end()

    def test_autocreate_mapping(self, collection, configurator):
        # This registers a representer (factory) and creates a mapping for
        # the collection.
        coll_rpr = as_representer(collection, CsvMime)
        mb = next(iter(collection))
        # This auto-creates a mapping for the member.
        coll_rpr.resource_to_data(mb)
        rpr_reg = configurator.get_registered_utility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(CsvMime)
        mp_before = mp_reg.find_mapping(type(mb))
        # This registers a representer (factory) for the member and finds
        # the previously created mapping for the member.
        as_representer(mb, CsvMime)
        mp_after = mp_reg.find_mapping(type(mb))
        assert mp_before is mp_after


class TestAttributes(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_defaults(self):
        rc_attr = get_resource_class_attribute(MyEntityMember, 'number')
        mp_attr = MappedAttribute(rc_attr)
        assert mp_attr.repr_name == rc_attr.entity_attr
        with pytest.raises(AttributeError):
            getattr(mp_attr, 'foo')
        assert str(mp_attr).startswith(mp_attr.attr_type.__name__)

    def test_ignore(self):
        rc_attr = get_resource_class_attribute(MyEntityMember, 'number')
        mp_attr = MappedAttribute(rc_attr,
                                  options={IGNORE_OPTION:False})
        assert getattr(mp_attr, IGNORE_OPTION) is False

    def test_clone(self):
        rc_attr = get_resource_class_attribute(MyEntityMember, 'number')
        mp_attr = MappedAttribute(rc_attr)
        mp_attr_clone = mp_attr.clone()
        assert mp_attr.options == mp_attr_clone.options
        assert mp_attr.name == mp_attr_clone.name
        assert mp_attr.kind == mp_attr_clone.kind
        assert mp_attr.value_type == mp_attr_clone.value_type
        assert mp_attr.entity_name == mp_attr_clone.entity_name
        assert mp_attr.cardinality == mp_attr_clone.cardinality


class TestCsvData(object):
    def test_methods(self):
        csvd0 = CsvData()
        od1 = OrderedDict()
        od1['foo'] = 'foo'
        od1['bar'] = 1
        csvd1 = CsvData(od1)
        csvd0.expand(csvd1)
        csvd0.fields == ['foo', 'bar'] # pylint: disable=W0104
        csvd0.data == [['foo', 1]] # pylint: disable=W0104
        od2 = OrderedDict()
        od2['foo'] = 'foo1'
        od2['bar'] = 2
        csvd2 = CsvData(od2)
        csvd0.append(csvd2)
        assert csvd0.fields == ['foo', 'bar']
        assert csvd0.data == [['foo', 1], ['foo1', 2]]


class _TestRepresenter(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_member_representer(self, member_representer, collection,
                                monkeypatch):
        mb = next(iter(collection))
        rpr_str = member_representer.to_string(mb)
        mb_reloaded = member_representer.from_string(rpr_str)
        assert mb.id == mb_reloaded.id
        # Check unicode handling.
        mb.text = u'h\xfclfe'
        bytes_val = bytes_(mb.text, encoding=member_representer.encoding)
        attr = get_domain_class_attribute(MyEntity, 'text')
        def test(member, exp_val):
            rpr_text = member_representer.to_string(member)
            assert isinstance(rpr_text, text_type)
            rpr_bytes = member_representer.to_bytes(member)
            assert isinstance(rpr_bytes, binary_type)
            mb_reloaded_str = member_representer.from_string(rpr_text)
            assert isinstance(mb_reloaded_str.text, attr.attr_type)
            assert mb_reloaded_str.text == exp_val
            mb_reloaded_bytes = member_representer.from_bytes(rpr_bytes)
            assert isinstance(mb_reloaded_bytes.text, attr.attr_type)
            assert mb_reloaded_bytes.text == exp_val
            #
            de = member_representer.resource_to_data(member)
            assert isinstance(de, MemberDataElement)
            rpr_data_bytes = member_representer.data_to_bytes(de)
            assert isinstance(rpr_data_bytes, binary_type)
        # In PY3, the attr type will be text, in PY2 bytes.
        if not issubclass(attr.attr_type, binary_type):
            monkeypatch.setattr(attr, 'attr_type', binary_type)
        test(mb, bytes_val)
        # In PY3, the attr type will be text, in PY2 bytes.
        if not issubclass(attr.attr_type, text_type):
            monkeypatch.setattr(attr, 'attr_type', text_type)
        test(mb, mb.text)

    def _test_with_defaults(self, representer, collection, check_string,
                            do_roundtrip=True):
        self._test_rpr(representer, collection, None, check_string,
                       self._check_nested_member if do_roundtrip else None)

    def _test_with_collection_link(self, representer, collection, check_string,
                                   do_roundtrip=True):
        attribute_options = {('children',):{IGNORE_OPTION:False,
                                            WRITE_AS_LINK_OPTION:True}}
        self._test_rpr(representer, collection, attribute_options,
                       check_string,
                    self._check_nested_collection if do_roundtrip else None)

    def _test_with_member_expanded(self, representer, collection,
                                   check_string, do_roundtrip=True):
        attribute_options = {('parent',):{WRITE_AS_LINK_OPTION:False}}
        self._test_rpr(representer, collection, attribute_options,
                       check_string,
                       self._check_nested_member if do_roundtrip else None)

    def _test_with_collection_expanded(self, representer, collection,
                                       check_string, do_roundtrip=True):
        attribute_options = {('children',):{IGNORE_OPTION:False,
                                            WRITE_AS_LINK_OPTION:False}}
        self._test_rpr(representer, collection, attribute_options,
                       check_string,
                    self._check_nested_collection if do_roundtrip else None)

    def _test_with_two_collections_expanded(self, representer, collection,
                                            check_string, do_roundtrip=True):
        attribute_options = \
            {('children',):{IGNORE_OPTION:False,
                            WRITE_AS_LINK_OPTION:False},
             ('children', 'children'):{IGNORE_OPTION:False,
                                       WRITE_AS_LINK_OPTION:False},
             }
        self._test_rpr(representer, collection, attribute_options,
                       check_string,
                    self._check_nested_collection if do_roundtrip else None)

    def _test_rpr(self, representer, collection, attribute_options,
                  str_checker, coll_checker):
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            rpr_str = representer.to_string(collection)
            assert len(rpr_str) > 0
            if not str_checker is None:
                str_checker(rpr_str)
            if not coll_checker is None:
                reloaded_coll = representer.from_string(rpr_str)
                assert len(reloaded_coll) == len(collection)
                coll_checker(reloaded_coll)

    def _check_id(self, collection):
        assert next(iter(collection)).id == next(iter(collection).id)

    def _check_nested_member(self, collection):
        assert next(iter(collection)).parent.id \
                        == next(iter(collection)).parent.id

    def _check_nested_collection(self, collection):
        assert next(iter(next(iter(collection)).children)).id \
                        == next(iter(next(iter(collection)).children)).id


class TestJsonRepresenter(_TestRepresenter):
    content_type = JsonMime

    def test_json_with_defaults(self, representer, collection):
        def check_string(rpr_str):
            assert rpr_str.startswith('[{') and rpr_str.endswith('}]')
        self._test_with_defaults(representer, collection, check_string)

    def test_json_with_collection_link(self, representer, collection):
        def check_string(rpr_str):
            assert rpr_str.find('my-entity-children/?q=parent') != -1
        self._test_with_collection_link(representer, collection, check_string)

    def test_json_with_member_expanded(self, representer, collection):
        def check_string(rpr_str):
            assert rpr_str.find('parent_text') != -1
        self._test_with_member_expanded(representer, collection, check_string)

    def test_json_with_two_collections_expanded(self, representer, collection):
        self._test_with_two_collections_expanded(representer, collection, None)

    def test_json_data_tree_traverser(self, configurator):
        configurator.begin()
        try:
            rpr_reg = configurator.registry.queryUtility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(JsonMime)
            default_mp = mp_reg.find_or_create_mapping(MyEntityMember)
            attr_opts = {('parent',):{WRITE_AS_LINK_OPTION:False}}
            mp = default_mp.clone(attribute_options=attr_opts)
            vst = DataElementBuilderRepresentationDataVisitor(mp)
            for json_data, exc_msg in ((object, 'Need dict (member),'),
                                       ({'parent':
                                            {'__jsonclass__':'http://foo.org'}},
                                        'Expected data for'),):
                trv = JsonDataTreeTraverser(json_data, mp)
                with pytest.raises(ValueError) as cm:
                    trv.run(vst)
                assert str(cm.value).startswith(exc_msg)
        finally:
            configurator.end()


class TestCsvRepresenter(_TestRepresenter):
    content_type = CsvMime

    def test_csv_with_defaults(self, representer, collection):
        def check_string(rpr_str):
            lines = rpr_str.strip().split(os.linesep)
            assert len(lines) == 3
            assert lines[0] == '"id","parent","text",' \
                               '"text_rc","number","date_time",' \
                               '"parent_text"'
            assert lines[1][0] == '0'
            assert lines[2][0] == '1'
            row_data = lines[1].split(',')
            # By default, members are represented as links.
            assert row_data[1].find('my-entity-parents/0/') != -1
            # By default, collections are not processed.
            assert row_data[-1] == '"TEXT"'
        self._test_with_defaults(representer, collection, check_string)

    def test_csv_with_collection_link(self, representer, collection):
        def check_string(rpr_str):
            lines = rpr_str.split(os.linesep)
            row_data = lines[1].split(',')
            # Now, the collection should be a URL.
            assert row_data[2].find('my-entity-children/?q=parent') != -1
        self._test_with_collection_link(representer, collection, check_string)

    def test_csv_with_member_expanded(self, representer, collection):
        def check_string(rpr_str):
            lines = rpr_str.split(os.linesep)
            assert lines[0] == '"id","parent.id","parent.text",' \
                               '"parent.text_rc","text",' \
                               '"text_rc","number","date_time",' \
                               '"parent_text"'
            row_data = lines[1].split(',')
            # Second field should be the "parent.id" and contain '0'.
            assert row_data[1] == '0'
        # Ensure unique representation names for nested attributes.
        attribute_options = {
            ('parent', 'id') : {REPR_NAME_OPTION:'parent.id'},
            ('parent', 'text') : {REPR_NAME_OPTION:'parent.text'},
            ('parent', 'text_rc') : {REPR_NAME_OPTION:'parent.text_rc'},
             }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            self._test_with_member_expanded(representer, collection,
                                            check_string)

    def test_csv_with_collection_expanded(self, representer, collection):
        def check_string(rpr_str):
            lines = rpr_str.split(os.linesep)
            assert lines[0] == '"id","parent","children.id",' \
                               '"children.text","children.text_rc",' \
                               '"text","text_rc","number","date_time",' \
                               '"parent_text"'
            row_data = lines[1].split(',')
            # Fourth field should now be "children.id" and contain 0.
            assert row_data[2] == '0'
            # Fifth field should be "children.text" and contain "TEXT".
            assert row_data[3] == '"TEXT"'
        # Ensure unique representation names for nested attributes.
        attribute_options = {
            ('children', 'id') : {REPR_NAME_OPTION:'children.id'},
            ('children', 'parent') : {IGNORE_OPTION:True},
            ('children', 'text') : {REPR_NAME_OPTION:'children.text'},
            ('children', 'text_rc') : {REPR_NAME_OPTION:'children.text_rc'},
             }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            self._test_with_collection_expanded(representer, collection,
                                                check_string)

    def test_csv_collection_to_data_roundtrip(self, representer, collection):
        attribute_options = {('parent',):{IGNORE_OPTION:True, },
                             ('parent_text',):{IGNORE_OPTION:True, }}
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            data_el = representer.resource_to_data(collection)
            loaded_coll = representer.resource_from_data(data_el)
            assert next(iter(loaded_coll)).parent is None

    def test_csv_member_to_data_roundtrip_in_place(self, representer,
                                                   collection):
        mb = collection['0']
        parent = mb.parent
        data_el = representer.resource_to_data(mb)
        representer.resource_from_data(data_el, resource=mb)
        assert mb.parent.get_entity() is parent.get_entity()

    def test_csv_with_two_collections_expanded(self, representer, collection):
        def check_string(rpr_str):
            lines = rpr_str.split(os.linesep)
            row_data = lines[1].split(',')
            #  field should be "children.children.id".
            assert row_data[4] == '0'
        self._test_with_two_collections_expanded(representer, collection,
                                                 check_string,
                                                 do_roundtrip=False)

    def test_csv_unicode_repr_name(self, representer, collection):
        attribute_options = \
            {('children',):{IGNORE_OPTION:True, },
             ('parent',):{IGNORE_OPTION:True, },
             ('parent_text',):{IGNORE_OPTION:True, },
             ('text',):{REPR_NAME_OPTION:u'custom_text'}
             }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            data = representer.resource_to_data(collection)
            rpr_str = representer.data_to_string(data)
            lines = rpr_str.split(os.linesep)
            assert lines[0] == '"id","custom_text","text_rc","number","date_time"'

    def test_csv_data_from_string(self, representer):
        csv_invalid_field = '"id","text","number","foo"\n0,"abc",0,"xyz"'
        with pytest.raises(ValueError) as cm:
            representer.data_from_string(csv_invalid_field)
        assert str(cm.value).startswith('Invalid field')
        csv_invalid_row_length = '"id","text","number"\n0,"abc",0,"xyz",5'
        with pytest.raises(ValueError) as cm:
            representer.data_from_string(csv_invalid_row_length)
        assert str(cm.value).startswith('Invalid row length')
        csv_value_for_ignored_field = '"id","children"\n' \
                                      '0,"http://0.0.0.0/my-entity-parents/0"'
        with pytest.raises(ValueError) as cm:
            representer.data_from_string(csv_value_for_ignored_field)
        assert str(cm.value).startswith('Value for attribute')
        csv_invalid_link = '"id","parent"\n0,"my-entity-parents/0"'
        with pytest.raises(ValueError) as cm:
            representer.data_from_string(csv_invalid_link)
        assert str(cm.value).startswith('Value for nested attribute')

    def test_csv_with_two_collections_expanded_fails(self, representer):
        attribute_options = {
            ('children',) : {IGNORE_OPTION:False},
            ('children', 'id') : {REPR_NAME_OPTION:'children.id'},
            ('children', 'children',) : {IGNORE_OPTION:False},
            ('children', 'children', 'id') :
                            {REPR_NAME_OPTION:'children.children.id'},
             }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            csv_two_colls_expanded = \
                    '"id","children.id","children.children.id"\n0,0,0'
            with pytest.raises(ValueError) as cm:
                representer.data_from_string(csv_two_colls_expanded)
            assert str(cm.value).startswith('All but one nested collection')

    def test_csv_with_multiple_nested_members(self, representer):
        attribute_options = {
            ('children',) : {IGNORE_OPTION:False},
            ('children', 'id') : {REPR_NAME_OPTION:'children.id'},
            }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            csv_multiple_nested_members = '"id","children.id"\n0,0\n0,1'
            data_el = representer.data_from_string(csv_multiple_nested_members)
            assert len(data_el) == 1
            mb_el = data_el.members[0]
            assert mb_el.data['children'].members[1].data['children.id'] == 1

    def test_csv_none_attribute_value(self, representer, collection):
        ent = collection['0'].get_entity()
        ent.text = None
        def check_string(rpr_str):
            lines = rpr_str.strip().split(os.linesep)
            assert len(lines) == 3
            # Make sure the header is correct.
            assert lines[0] == '"id","parent","text",' \
                               '"text_rc","number","date_time",' \
                               '"parent_text"'
            # None value represented as the empty string.
            assert lines[1].split(',')[2] == '""'
            assert lines[2].split(',')[2] == '"too1"'
        self._test_with_defaults(representer, collection, check_string)


class TestXmlRepresenter(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'
    content_type = XmlMime

    def test_xml_with_defaults(self, representer, collection):
        rpr_str = representer.to_string(collection)
        assert rpr_str.find('<ent:myentityparent id="0">') != -1
        de = representer.resource_to_data(collection)
        assert isinstance(de, CollectionDataElement)
        rpr_data_bytes = representer.data_to_bytes(de)
        assert isinstance(rpr_data_bytes, binary_type)

    def test_xml_roundtrip(self, representer, collection):
        attribute_options = \
                {('text_rc',):{IGNORE_OPTION:True},
                 ('parent_text',):{IGNORE_OPTION:True},
                 ('children',):{IGNORE_OPTION:False,
                                WRITE_AS_LINK_OPTION:True},
                 }
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            data = representer.resource_to_data(collection)
            assert len(data) == 2
            rpr_str = representer.data_to_string(data)
            reloaded_coll = representer.from_string(rpr_str)
            assert len(reloaded_coll) == 2

    def test_id_attr(self, member_mapping):
        id_attr = member_mapping.get_attribute_map()['id']
        de = member_mapping.data_element_class.create()
        assert de.get_terminal(id_attr) is None

    def test_terminal_attr(self, collection, member_mapping):
        mb = next(iter(collection))
        text_attr = member_mapping.get_attribute_map()['text']
        de = member_mapping.map_to_data_element(mb)
        assert de.get_terminal(text_attr) == mb.text
        mb.text = None
        de1 = member_mapping.map_to_data_element(mb)
        assert de1.get_terminal(text_attr) is None

    @pytest.mark.parametrize(
                        'attribute_options,check_for_link',
                        [({('parent',):{WRITE_AS_LINK_OPTION:False}}, False),
                         ({('parent',):{WRITE_AS_LINK_OPTION:True}}, True)])
    def test_data(self, collection, representer, attribute_options,
                  check_for_link):
        mb = next(iter(collection))
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            de = representer.resource_to_data(mb)
            assert list(de.data.keys()) \
                    == ['id', 'myentityparent', 'text', 'number', 'date_time']
            assert de.get_attribute('number') == 1
            parent = de.get_attribute('myentityparent')
            with pytest.raises(AttributeError):
                dummy = de.get_attribute('foo')
            with pytest.raises(AttributeError):
                de.set_attribute('foo', 'bar')
            if not check_for_link:
                assert IMemberDataElement.providedBy(parent) # pylint:disable=E1101
            else:
                assert ILinkedDataElement.providedBy(parent) # pylint:disable=E1101

    def test_create(self, mapping):
        xml_tag = mapping.configuration.get_option(XML_TAG_OPTION)
        xml_ns = mapping.configuration.get_option(XML_NAMESPACE_OPTION)
        de = mapping.data_element_class.create()
        assert de.tag == '{%s}%s' % (xml_ns, xml_tag)
        assert de.nsmap[None] == xml_ns

    def test_create_no_namespace(self, representer, mapping):
        xml_tag = mapping.configuration.get_option(XML_TAG_OPTION)
        options = {XML_NAMESPACE_OPTION:None}
        with representer.with_updated_configuration(options=options):
            de = mapping.data_element_class.create()
            assert de.tag == xml_tag

    def test_create_with_attr_namespace(self, collection, representer,
                                        mapping):
        mb = next(iter(collection))
        ns = 'foo'
        attribute_options = {('parent',):{NAMESPACE_MAPPING_OPTION:ns,
                                          WRITE_AS_LINK_OPTION:False}}
        with representer.with_updated_configuration(attribute_options=
                                                        attribute_options):
            attr = mapping.get_attribute_map()['parent']
            assert attr.namespace == ns
            de = representer.resource_to_data(mb)
            parent_de = de.get_nested(attr)
            assert parent_de.tag.startswith('{%s}' % ns)

    def test_create_with_attr_no_namespace(self, member, new_configurator):
        new_configurator.begin()
        try:
            rpr_reg = \
                new_configurator.get_registered_utility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(XmlMime)
            ns = None
            options = {XML_NAMESPACE_OPTION:ns}
            mp = mp_reg.find_or_create_mapping(MyEntityParentMember)
            with mp.with_updated_configuration(options=options):
                # We need to do this so the namespace lookup is rebuilt.
                mp_reg.set_mapping(mp)
                attr = mp.get_attribute('id')
                assert attr.namespace == ns
                de = mp.map_to_data_element(member.parent)
                assert list(de.data.keys()) == ['id', 'text', 'text_rc']
                assert de.tag.find('{') == -1
        finally:
            new_configurator.end()

    def test_create_no_tag_raises_error(self, mapping):
        options = {XML_TAG_OPTION:None}
        with mapping.with_updated_configuration(options=options):
            with pytest.raises(ValueError):
                mapping.data_element_class.create()

    def test_create_link(self, collection, configurator):
        rpr_reg = configurator.get_registered_utility(IRepresenterRegistry)
        mp_reg = rpr_reg.get_mapping_registry(XmlMime)
        mp = mp_reg.find_mapping(Link)
        de = mp.data_element_class.create_from_resource(collection)
        link_el = next(de.iterchildren())
        assert link_el.get_kind() == RESOURCE_KINDS.COLLECTION
        assert link_el.get_relation().find('myentity') != -1
        assert link_el.get_title().startswith('Collection of')
        assert link_el.get_id() is None

    def test_create_link_from_non_resource_raises_error(self, configurator):
        configurator.begin()
        try:
            non_rc = NonResource()
            rpr_reg = configurator.get_registered_utility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(XmlMime)
            mp = mp_reg.find_mapping(Link)
            with pytest.raises(ValueError):
                mp.data_element_class.create_from_resource(non_rc)
        finally:
            configurator.end()

    def test_invalid_xml(self, representer):
        with pytest.raises(SyntaxError) as cm:
            representer.from_string('<murks/>')
        exc_msg = 'Could not parse XML document for schema'
        assert str(cm.value).find(exc_msg) != -1

    def test_no_xml_string_as_schema(self, representer):
        options = {XML_SCHEMA_OPTION:'everest:tests/complete_app/NoXml.xsd'}
        with representer.with_updated_configuration(options=options):
            with pytest.raises(SyntaxError) as cm:
                representer.from_string('')
            exc_msg = 'Could not parse XML schema'
            assert str(cm.value).find(exc_msg) != -1

    def test_no_schema_xml_string_as_schema(self, representer):
        options = \
            {XML_SCHEMA_OPTION:'everest:tests/complete_app/NoSchema.xsd'}
        with representer.with_updated_configuration(options=options):
            with pytest.raises(SyntaxError) as cm:
                representer.from_string('')
            exc_msg = 'Invalid XML schema'
            assert str(cm.value).find(exc_msg) != -1


class TestAtomRepresenter(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_rpr.zcml'
    content_type = AtomMime

    def test_atom_collection(self, collection, representer):
        rpr_str1 = representer.to_string(collection)
        assert rpr_str1.find('<feed xmlns:') != -1
        collection.slice = slice(0, 1)
        filter_spec_fac = get_filter_specification_factory()
        filter_spec = filter_spec_fac.create_equal_to('id', 0)
        collection.filter = filter_spec
        order_spec_fac = get_order_specification_factory()
        order_spec = order_spec_fac.create_ascending('id')
        collection.order = order_spec
        rpr_str2 = representer.to_string(collection)
        assert rpr_str2.find('<feed xmlns:') != -1

    def test_atom_member(self, collection, member_representer):
        mb = next(iter(collection))
        rpr_str = member_representer.to_string(mb)
        assert \
            rpr_str.find('<entry xmlns:ent="http://xml.test.org/tests"') != -1


class _TestRepresenterConfiguration(object):
    package_name = 'everest.tests.complete_app'
    content_type = CsvMime

    def test_configure_rpr_with_zcml(self, collection, representer):
        rpr_str = representer.to_string(collection)
        lines = rpr_str.split(os.linesep)
        fields = lines[0].split(',')
        # Parent should be inline.
        assert '"parent.id"' in fields
        chld_field_idx = fields.index('"children"')
        row_data = lines[1].split(',')
        # Collection should be a link.
        assert \
          row_data[chld_field_idx].find('my-entity-children/?q=parent') != -1


class TestRepresenterConfigurationOldStyleAttrs(
                                            _TestRepresenterConfiguration):
    config_file_name = 'configure_rpr_oldstyle_attrs.zcml'


class TestRepresenterConfigurationNoTypes(_TestRepresenterConfiguration):
    config_file_name = 'configure_rpr_no_types.zcml'


class TestRepresenterConfiguration(_TestRepresenterConfiguration):
    config_file_name = 'configure_rpr.zcml'

    def test_custom_config(self):
        class MyRepresenterConfiguration(RepresenterConfiguration):
            _default_config_options = dict(
                list(RepresenterConfiguration._default_config_options.items())
                     + [('foo', None)])
        foo_val = 1
        rpr_cnf = MyRepresenterConfiguration(
                        options=dict(foo=foo_val),
                        attribute_options=dict(foo={IGNORE_OPTION:True}))
        rpr_cnf_cp = rpr_cnf.copy()
        assert rpr_cnf.get_option('foo') == foo_val
        assert rpr_cnf_cp.get_option('foo') == foo_val
        for key in (('foo',), ['foo']):
            assert rpr_cnf.get_attribute_option(key, IGNORE_OPTION) is True
            assert rpr_cnf_cp.get_attribute_option(key, IGNORE_OPTION) is True

    def test_non_existing_attribute_fails(self, resource_repo):
        coll = resource_repo.get_collection(IMyEntity)
        ent = MyEntity(id=1)
        mb = coll.create_member(ent)
        rpr = as_representer(mb, CsvMime)
        attr_name = 'invalid'
        attribute_options = {(attr_name,):{IGNORE_OPTION:False,
                                           WRITE_AS_LINK_OPTION:True}}
        with pytest.raises(AttributeError) as cm:
            with rpr.with_updated_configuration():
                rpr.configure(attribute_options=attribute_options)
        msg = 'Trying to configure non-existing resource attribute "%s"' \
              % attr_name
        assert str(cm.value)[:len(msg)] == msg

    def test_configure_existing(self, new_configurator):
        new_configurator.begin()
        try:
            foo_namespace = 'http://bogus.org/foo'
            foo_prefix = 'foo'
            my_options = {XML_NAMESPACE_OPTION:foo_namespace,
                          XML_PREFIX_OPTION:foo_prefix}
            my_attribute_options = {('parent',):{IGNORE_OPTION:True}, }
            new_configurator.add_resource_representer(
                                        MyEntityMember, XmlMime,
                                        options=my_options,
                                        attribute_options=
                                                    my_attribute_options)
            rpr_reg = \
                new_configurator.get_registered_utility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(XmlMime)
            mp = mp_reg.find_mapping(MyEntityMember)
            assert mp.configuration.get_option(XML_NAMESPACE_OPTION) \
                    == foo_namespace
            assert mp.configuration.get_attribute_option(('parent',),
                                                         IGNORE_OPTION) \
                    is True
            with pytest.raises(ValueError):
                with mp.with_updated_configuration():
                    mp.configuration.set_option('nonsense', True)
            with pytest.raises(ValueError):
                with mp.with_updated_configuration():
                    mp.configuration.set_attribute_option(('parent',),
                                                          'nonsense', True)
        finally:
            new_configurator.end()

    def test_configure_derived(self, new_configurator):
        new_configurator.begin()
        try:
            new_configurator.add_resource(IDerived, DerivedMyEntityMember,
                                          DerivedMyEntity,
                                          collection_root_name=
                                                  'my-derived-entities',
                                          expose=False)
            new_configurator.add_resource_representer(DerivedMyEntityMember,
                                                      XmlMime)
            rpr_reg = \
                new_configurator.get_registered_utility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(XmlMime)
            mp = mp_reg.find_mapping(DerivedMyEntityMember)
            assert mp.data_element_class.mapping is mp
            assert mp.configuration.get_option(XML_TAG_OPTION) == 'myentity'
        finally:
            new_configurator.end()

    def test_configure_derived_with_options(self, new_configurator):
        new_configurator.begin()
        try:
            new_configurator.add_resource(IDerived, DerivedMyEntityMember,
                                          DerivedMyEntity,
                                          collection_root_name=
                                                    'my-derived-entities',
                                          expose=False)
            foo_namespace = 'http://bogus.org/foo'
            bogus_prefix = 'foo'
            my_options = {XML_NAMESPACE_OPTION:foo_namespace,
                          XML_PREFIX_OPTION:bogus_prefix}
            new_configurator.add_resource_representer(DerivedMyEntityMember,
                                                      XmlMime,
                                                      options=my_options)
            rpr_reg = \
                new_configurator.get_registered_utility(IRepresenterRegistry)
            mp_reg = rpr_reg.get_mapping_registry(XmlMime)
            mp = mp_reg.find_mapping(DerivedMyEntityMember)
            assert mp.data_element_class.mapping is mp
            assert mp.configuration.get_option(XML_NAMESPACE_OPTION) \
                    == foo_namespace
            assert mp.configuration.get_option(XML_PREFIX_OPTION) \
                    == bogus_prefix
            orig_mp = mp_reg.find_mapping(MyEntityMember)
            assert not orig_mp is mp
            assert orig_mp.data_element_class.mapping is orig_mp
            assert orig_mp.configuration.get_option(XML_NAMESPACE_OPTION) \
                    != foo_namespace
            assert orig_mp.configuration.get_option(XML_PREFIX_OPTION) \
                    != bogus_prefix
        finally:
            new_configurator.end()


class TestUpdateResourceFromData(object):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_update_nested_member_from_data(self, resource_repo):
        # Set up member that does not have a parent.
        coll = resource_repo.get_collection(IMyEntity)
        ent = MyEntity(id=1)
        mb = coll.create_member(ent)
        # Set up second member with same ID that does have a parent.
        tmp_coll = create_staging_collection(IMyEntity)
        parent = MyEntityParent(id=0)
        upd_ent = MyEntity(id=1, parent=parent)
        upd_mb = tmp_coll.create_member(upd_ent)
        rpr = as_representer(mb, CsvMime)
        attribute_options = {('parent',):{WRITE_AS_LINK_OPTION:False}}
        with rpr.with_updated_configuration(attribute_options=
                                                attribute_options):
            de = rpr.resource_to_data(upd_mb)
            mb.update(de)
            assert mb.parent.id == parent.id

    def test_update_nested_collection_from_data(self, resource_repo):
        # Set up member that has one child.
        coll = resource_repo.get_collection(IMyEntity)
        ent = MyEntity(id=1)
        child0 = MyEntityChild(id=0)
        ent.children.append(child0)
        mb = coll.create_member(ent)
        # Set up another member with two children with different IDs.
        tmp_coll = create_staging_collection(IMyEntity)
        upd_ent = MyEntity(id=1)
        child1 = MyEntityChild(id=1)
        child1.parent = upd_ent
        child2 = MyEntityChild(id=2)
        child2.parent = upd_ent
        upd_ent.children.append(child1)
        upd_ent.children.append(child2)
        upd_mb = tmp_coll.create_member(upd_ent)
        rpr = as_representer(mb, CsvMime)
        attribute_options = {('children',):{IGNORE_OPTION:False,
                                            WRITE_AS_LINK_OPTION:False}, }
        with rpr.with_updated_configuration(attribute_options=
                                                    attribute_options):
            de = rpr.resource_to_data(upd_mb)
            mb.update(de)
            assert set([mb.id for mb in mb.children]) == set([1, 2])


# pylint: disable=W0232
class IDerived(Interface):
    pass
# pylint: enable=W0232


class DerivedMyEntity(MyEntity):
    pass


class DerivedMyEntityMember(MyEntityMember):
    pass


class NonResource(object):
    pass



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 17, 2014.
"""
import glob
import os
import shutil

from pyramid.testing import setUp as set_up_testing
from pyramid.testing import tearDown as tear_down_testing
from pyramid.threadlocal import get_current_registry
import pytest

from everest.configuration import Configurator
from everest.tests import simple_app as package


__docformat__ = 'reStructuredText en'
__all__ = ['simple_config',
           ]


@pytest.fixture
def simple_config(request):
    set_up_testing()
    reg = get_current_registry()
    config = Configurator(registry=reg, package=package)
    config.setup_registry()
    request.addfinalizer(tear_down_testing)
    return config


@pytest.fixture
def data_dir(request):
    pkg_name = getattr(request.cls, 'package_name', None)
    return os.path.join(os.path.dirname(__file__), pkg_name.split('.')[-1],
                        'data')


@pytest.yield_fixture
def resource_repo_with_data(data_dir, resource_repo): # pylint: disable=W0621
    with CopyDataFilesContextManager(data_dir):
        yield resource_repo


class CopyDataFilesContextManager(object):
    def __init__(self, data_directory):
        self.__data_dir = data_directory

    def __enter__(self):
        orig_data_dir = os.path.join(self.__data_dir, 'original')
        for fn in glob.glob1(orig_data_dir, "*.csv"):
            shutil.copy(os.path.join(orig_data_dir, fn), self.__data_dir)
        return self.__data_dir

    def __exit__(self, ext_type, value, tb):
        for fn in glob.glob1(self.__data_dir, '*.csv'):
            os.unlink(os.path.join(self.__data_dir, fn))



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 4, 2011.
"""
from datetime import datetime
from everest.querying.filterparser import parse_filter
from everest.querying.operators import ENDS_WITH
from everest.querying.operators import EQUAL_TO
from everest.querying.operators import GREATER_THAN
from everest.querying.operators import STARTS_WITH
from everest.querying.specifications import ConjunctionFilterSpecification
from everest.querying.specifications import DisjunctionFilterSpecification
from everest.querying.specifications import ValueEqualToFilterSpecification
from everest.querying.specifications import ValueGreaterThanFilterSpecification
from everest.querying.specifications import ValueLessThanFilterSpecification
from everest.querying.specifications import ValueStartsWithFilterSpecification
from everest.testing import TestCaseWithConfiguration
from pyparsing import ParseException

__docformat__ = 'reStructuredText en'
__all__ = ['QueryParserTestCase',
           ]


class QueryParserTestCase(TestCaseWithConfiguration):
    parser = None

    def set_up(self):
        TestCaseWithConfiguration.set_up(self)
        self.parser = parse_filter

    def test_no_criterion_query(self):
        expr = ''
        self.assert_raises(ParseException, self.parser, expr)

    def test_one_criterion_query(self):
        expr = 'name:equal-to:"Nikos"'
        result = self.parser(expr)
        self.assert_true(isinstance(result, ValueEqualToFilterSpecification))
        self.assert_equal(result.attr_name, 'name')
        self.assert_equal(result.operator, EQUAL_TO)
        self.assert_equal(result.attr_value, 'Nikos')

    def test_one_criterion_query_with_many_values(self):
        expr = 'name:equal-to:"Nikos","Oliver","Andrew"'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec,
                                    DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.right_spec,
                                    ValueEqualToFilterSpecification))
        self.assert_equal(result.left_spec.left_spec.attr_name, 'name')
        self.assert_equal(result.left_spec.left_spec.operator, EQUAL_TO)
        self.assert_equal(result.left_spec.left_spec.attr_value, 'Nikos')
        self.assert_equal(result.left_spec.right_spec.attr_name, 'name')
        self.assert_equal(result.left_spec.right_spec.operator, EQUAL_TO)
        self.assert_equal(result.left_spec.right_spec.attr_value, 'Oliver')
        self.assert_equal(result.right_spec.attr_name, 'name')
        self.assert_equal(result.right_spec.operator, EQUAL_TO)
        self.assert_equal(result.right_spec.attr_value, 'Andrew')

    def test_mixed_criteria_query(self):
        expr = 'name:starts-with:"Ni" OR name:ends-with:"kos" ' \
               'AND age:equal-to:34'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec,
                                    ValueStartsWithFilterSpecification))
        self.assert_true(isinstance(result.right_spec,
                                    ConjunctionFilterSpecification))
        self.assert_equal(result.left_spec.attr_name, 'name')
        self.assert_equal(result.right_spec.left_spec.attr_name, 'name')
        self.assert_equal(result.right_spec.right_spec.attr_name, 'age')

    def test_mixed_criteria_query_with_parens(self):
        expr = '(name:starts-with:"Ni" AND name:ends-with:"kos") ' \
               'OR age:equal-to:34'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec,
                                    ConjunctionFilterSpecification))
        self.assert_true(isinstance(result.right_spec,
                                    ValueEqualToFilterSpecification))
        self.assert_equal(result.left_spec.left_spec.attr_name, 'name')
        self.assert_equal(result.left_spec.right_spec.attr_name, 'name')
        self.assert_equal(result.right_spec.attr_name, 'age')

    def test_nested_criteria_query(self):
        expr0 = '(name:starts-with:"Ni" AND ' \
               ' (name:ends-with:"kos" OR age:equal-to:34))'
        result0 = self.parser(expr0)
        self.assert_true(isinstance(result0, ConjunctionFilterSpecification))
        expr1 = '((name:starts-with:"Ni" AND name:ends-with:"kos") ' \
               'OR age:equal-to:34))'
        result1 = self.parser(expr1)
        self.assert_true(isinstance(result1, DisjunctionFilterSpecification))

    def test_multiple_criterion_query(self):
        def _test(expr):
            result = self.parser(expr)
            self.assert_true(isinstance(result,
                                        ConjunctionFilterSpecification))
            spec1 = result.left_spec
            spec2 = result.right_spec
            self.assert_equal(spec1.attr_name, 'name')
            self.assert_equal(spec1.operator, STARTS_WITH)
            self.assert_equal(spec1.attr_value, 'Ni')
            self.assert_equal(spec2.attr_name, 'name')
            self.assert_equal(spec2.operator, ENDS_WITH)
            self.assert_equal(spec2.attr_value, 'kos')
        crits = ['name:starts-with:"Ni"', 'name:ends-with:"kos"']
        for expr in ('~'.join(crits),
                     ' AND '.join(crits),
                     '(%s)' % ' AND '.join(crits)
                     ):
            _test(expr)

    def test_one_criterion_query_with_integers(self):
        expr = 'age:equal-to:34,44'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec,
                                    ValueEqualToFilterSpecification))
        self.assert_true(isinstance(result.right_spec,
                                    ValueEqualToFilterSpecification))
        self.assert_equal(result.left_spec.attr_name, 'age')
        self.assert_equal(result.left_spec.operator, EQUAL_TO)
        self.assert_equal(result.left_spec.attr_value, 34)
        self.assert_equal(result.right_spec.attr_name, 'age')
        self.assert_equal(result.right_spec.operator, EQUAL_TO)
        self.assert_equal(result.right_spec.attr_value, 44)

    def test_one_criterion_query_with_integer_scientific_format(self):
        expr = 'volume:greater-than:5e+05'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueGreaterThanFilterSpecification))
        self.assert_equal(result.attr_name, 'volume')
        self.assert_equal(result.operator, GREATER_THAN)
        self.assert_equal(result.attr_value, 500000)

    def test_one_criterion_query_with_floats(self):
        expr = 'cost:greater-than:3.14'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueGreaterThanFilterSpecification))
        self.assert_equal(result.attr_name, 'cost')
        self.assert_equal(result.operator, GREATER_THAN)
        self.assert_equal(result.attr_value, 3.14)

    def test_one_criterion_query_with_floats_scientific_format(self):
        expr = 'volume:greater-than:5.5e-05'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueGreaterThanFilterSpecification))
        self.assert_equal(result.attr_name, 'volume')
        self.assert_equal(result.operator, GREATER_THAN)
        self.assert_equal(result.attr_value, 5.5e-05)

    def test_one_criterion_query_with_floats_scientific_format_negative(self):
        expr = 'volume:greater-than:5e-05'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueGreaterThanFilterSpecification))
        self.assert_equal(result.attr_name, 'volume')
        self.assert_equal(result.operator, GREATER_THAN)
        self.assert_equal(result.attr_value, 5e-05)

    def _test_multiple_criterion_query_with_different_value_types(self, expr):
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ConjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec,
                                    ConjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec.left_spec,
                                    ConjunctionFilterSpecification))
        self.assert_true(isinstance(result.left_spec.left_spec.left_spec,
                                    DisjunctionFilterSpecification))
        self.assert_true(
            isinstance(result.left_spec.left_spec.left_spec.left_spec,
                       DisjunctionFilterSpecification))
        self.assert_true(isinstance(result.right_spec,
                                    DisjunctionFilterSpecification))
        spec1 = result.left_spec.left_spec.left_spec.left_spec.left_spec
        spec2 = result.left_spec.left_spec.left_spec.left_spec.right_spec
        spec3 = result.right_spec.left_spec
        spec4 = result.right_spec.right_spec
        self.assert_equal(spec1.attr_name, 'name')
        self.assert_equal(spec1.operator, STARTS_WITH)
        self.assert_equal(spec1.attr_value, 'Ni')
        self.assert_equal(spec2.attr_name, 'name')
        self.assert_equal(spec2.operator, STARTS_WITH)
        self.assert_equal(spec2.attr_value, 'Ol')
        self.assert_equal(spec3.attr_name, 'discount')
        self.assert_equal(spec3.operator, EQUAL_TO)
        self.assert_equal(spec3.attr_value, -20)
        self.assert_equal(spec4.attr_name, 'discount')
        self.assert_equal(spec4.operator, EQUAL_TO)
        self.assert_equal(spec4.attr_value, -30)

    def test_multiple_criterion_query_with_different_value_types(self):
        expr = 'name:starts-with:"Ni","Ol","An"~' \
               'age:equal-to:34,44,54~' \
               'phone-number:starts-with:1,2,3~' \
               'discount:equal-to:-20,-30~'
        self._test_multiple_criterion_query_with_different_value_types(expr)

    def test_multiple_criterion_query_with_misplaced_commas(self):
        expr = 'name:starts-with:"Ni", "Ol", ,"An"~' \
               'age:equal-to:34, 44, 54,~' \
               'phone-number:starts-with:,1,2,3~' \
               'discount:equal-to:,-20,,-30,'
        self._test_multiple_criterion_query_with_different_value_types(expr)

    def test_one_text_criterion_query_with_spaces(self):
        expr = 'name:equal-to:"Nikos Papagrigoriou"'
        result = self.parser(expr)
        self.assert_true(isinstance(result, ValueEqualToFilterSpecification))
        self.assert_equal(result.attr_value, 'Nikos Papagrigoriou')

    def test_invalid_criterion_with_only_comma_as_value(self):
        expr = 'name:equal-to:,'
        self.assert_raises(ValueError, self.parser, expr)

    def test_float_and_int(self):
        expr = 'age:less-than:12~height:less-than:5.2'
        result = self.parser(expr)
        self.assert_true(isinstance(result, ConjunctionFilterSpecification))
        self.assert_equal(result.left_spec.attr_value, 12)
        self.assert_true(result.right_spec.attr_value, 5.2)

    def test_valid_dotted_names(self):
        expr = 'user.age:less-than:12'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueLessThanFilterSpecification))
        self.assert_equal(result.attr_name, 'user.age')
        expr = 'user.address.street:equal-to:"Main"'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueEqualToFilterSpecification))
        self.assert_equal(result.attr_name, 'user.address.street')

    def test_invalid_dotted_names(self):
        expr = 'user.age.:less-than:12'
        self.assert_raises(ParseException, self.parser, expr)
        expr = '.user.age:less-than:12'
        self.assert_raises(ParseException, self.parser, expr)
        expr = 'user..age:less-than:12'
        self.assert_raises(ParseException, self.parser, expr)

    def test_valid_date(self):
        expr = 'birthday:equal-to:"1966-04-21T15:23:01Z"'
        result = self.parser(expr)
        self.assert_true(isinstance(result,
                                    ValueEqualToFilterSpecification))
        dt = result.attr_value
        self.assert_true(isinstance(dt, datetime))
        self.assert_equal(dt.year, 1966)
        self.assert_equal(dt.month, 4)
        self.assert_equal(dt.day, 21)
        self.assert_equal(dt.hour, 15)
        self.assert_equal(dt.minute, 23)
        self.assert_equal(dt.second, 1)

    def test_invalid_dates(self):
        def _check_expr(expr):
            result = self.parser(expr)
            self.assert_true(isinstance(result,
                                        ValueEqualToFilterSpecification))
            self.assert_false(isinstance(result.attr_value, datetime))
        # Violations of the regex.
        expr = 'birthday:equal-to:"19661-04-21T15:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"19661-041-21T15:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-211T15:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-21T151:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-21T15:231:00Z"'
        _check_expr(expr)
        # Violations of the allowed values.
        expr = 'birthday:equal-to:"1966-13-21T15:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-32T15:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-21T24:23:00Z"'
        _check_expr(expr)
        expr = 'birthday:equal-to:"1966-04-21T15:61:00Z"'
        _check_expr(expr)



"""
General attribute access functions.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 22, 2013.
"""
from everest.constants import CARDINALITY_CONSTANTS
from everest.constants import RESOURCE_ATTRIBUTE_KINDS

__docformat__ = 'reStructuredText en'
__all__ = ['get_attribute_cardinality',
           'is_terminal_attribute',
           ]


def is_terminal_attribute(attribute):
    """
    Checks if the given resource attribute is a terminal attribute.
    """
    return attribute.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL


def get_attribute_cardinality(attribute):
    """
    Returns the cardinality of the given resource attribute.

    :returns: One of the constants defined in
      :class:`evererst.constants.CARDINALITY_CONSTANTS`.
    :raises ValueError: If the given attribute is not a relation attribute
      (i.e., if it is a terminal attribute).
    """
    if attribute.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
        card = CARDINALITY_CONSTANTS.ONE
    elif attribute.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
        card = CARDINALITY_CONSTANTS.MANY
    else:
        raise ValueError('Can not determine cardinality for non-terminal '
                         'attributes.')
    return card



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 16, 2013.
"""

__docformat__ = 'reStructuredText en'
__all__ = ['TraversalPath',
           'TraversalPathNode',
           ]


class TraversalPathNode(object):
    """
    Value object representing a node in a traversal path.
    """
    def __init__(self, proxy, key, attribute, relation_operation):
        """
        :param proxy: Data traversal proxy for this node.
        :param attribute: Resource attribute for this node.
        :param relation_operation: Relation operation for this node. One of
          the constants defined in
          :class:`everest.constants.RELATION_OPERATIONS`.
        """
        self.proxy = proxy
        self.key = key
        self.attribute = attribute
        self.relation_operation = relation_operation


class TraversalPath(object):
    """
    Value object tracking a path taken by a data tree traverser.
    """
    def __init__(self, nodes=None):
        if nodes is None:
            nodes = []
        self.nodes = nodes
        self.__keys = set()

    def push(self, proxy, key, attribute, relation_operation):
        """
        Adds a new :class:`TraversalPathNode` constructed from the given
        arguments to this traversal path.
        """
        node = TraversalPathNode(proxy, key, attribute, relation_operation)
        self.nodes.append(node)
        self.__keys.add(key)

    def pop(self):
        """
        Removes the last traversal path node from this traversal path.
        """
        node = self.nodes.pop()
        self.__keys.remove(node.key)

    def __len__(self):
        return len(self.nodes)

    def __contains__(self, key):
        return key in self.__keys

    @property
    def parent(self):
        """
        Returns the proxy from the last node visited on the path, or `None`,
        if no node has been visited yet.
        """
        if len(self.nodes) > 0:
            parent = self.nodes[-1].proxy
        else:
            parent = None
        return parent

    @property
    def relation_operation(self):
        """
        Returns the relation operation from the last node visited on the
        path, or `None`, if no node has been visited yet.
        """
        if len(self.nodes) > 0:
            rel_op = self.nodes[-1].relation_operation
        else:
            rel_op = None
        return rel_op



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 1, 2011.
"""
from everest.entities.base import Entity
from pytz import timezone
import datetime

__docformat__ = 'reStructuredText en'
__all__ = ['MyEntity',
           'MyEntityChild',
           'MyEntityGrandchild',
           'MyEntityParent', ]


class _MyEntity(Entity):
    DEFAULT_TEXT = 'TEXT'

    text = None
    text_ent = None

    def __init__(self, text=None, text_ent=None, **kw):
        Entity.__init__(self, **kw)
        if text is None:
            text = self.DEFAULT_TEXT
        self.text = text
        if text_ent is None:
            text_ent = self.DEFAULT_TEXT
        self.text_ent = text_ent


class MyEntityParent(_MyEntity):

    child = None

    def __init__(self, child=None, **kw):
        _MyEntity.__init__(self, **kw)
        self.child = child


class MyEntity(_MyEntity):
    DEFAULT_NUMBER = 1
    DEFAULT_DATETIME = datetime.datetime(2012, 8, 29, 16, 20, 0,
                                         tzinfo=timezone('UTC'))

    parent = None
    children = None

    def __init__(self, parent=None, children=None, number=None,
                 date_time=None, **kw):
        _MyEntity.__init__(self, **kw)
        self.parent = parent
        if children is None:
            children = []
        self.children = children
        if number is None:
            number = self.DEFAULT_NUMBER
        self.number = number
        if date_time is None:
            date_time = self.DEFAULT_DATETIME
        self.date_time = date_time


class MyEntityChild(_MyEntity):

    parent = None
    children = None

    def __init__(self, parent=None, children=None, **kw):
        _MyEntity.__init__(self, **kw)
        self.parent = parent
        if children is None:
            children = []
        self.children = children


class MyEntityGrandchild(_MyEntity):

    parent = None

    def __init__(self, parent=None, **kw):
        _MyEntity.__init__(self, **kw)
        self.parent = parent



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Package setup file.

Created on Nov 3, 2011.
"""
import os
from distribute_setup import use_setuptools
use_setuptools()
from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup_requirements = []

install_requirements = \
    open(os.path.join(here, 'requirements.txt'), 'rU').readlines()

tests_requirements = install_requirements + [
    'pytest',
    'nose-cov',
    'pytz',
    'webtest',
    'mock'
    ]

setup(name='everest',
      version='1.0',
      description='everest',
      long_description=README,
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        ],
      author='F. Oliver Gathmann',
      author_email='fogathmann at gmail.com',
      license="MIT",
      url='https://github.com/cenix/everest',
      keywords='REST web wsgi pyramid',
      packages=find_packages(),
      package_data={'': ["*.zcml", "*.xsd"]},
      include_package_data=True,
      zip_safe=False,
      setup_requires=setup_requirements,
      install_requires=install_requirements,
      tests_require=tests_requirements,
      test_suite="everest,doc",
      entry_points="""\
      [nose.plugins.0.10]
      everest = everest.ini:EverestNosePlugin
      [pytest11]
      everest = everest.tests.fixtures
      """
      )



from everest.entities.base import Entity


class FooEntity(Entity):
    name = None
    def __init__(self, name=None, **kw):
        Entity.__init__(self, **kw)
        self.name = name

    @property
    def slug(self):
        if self.name is None:
            slug = str(self.id)
        else:
            slug = self.name
        return slug


class BarEntity(Entity):
    pass



"""
PATCH member view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 29, 2013.
"""
from everest.mime import XmlMime
from everest.representers.utils import UpdatedRepresenterConfigurationContext
from everest.representers.xml import XML_VALIDATE_OPTION
from everest.views.base import PutOrPatchResourceView


__docformat__ = 'reStructuredText en'
__all__ = ['PatchMemberView',
           ]


class PatchMemberView(PutOrPatchResourceView):
    """
    View for PATCH requests on member resources.

    The client sends a PATCH request to perform a partial update on a member
    resource. Note that the representation sent by the client typically is
    incomplete and possibly invalid (cf. PUT for replacing a member).
    """
    def _extract_request_data(self):
        rpr = self._get_request_representer()
        if rpr.content_type is XmlMime:
            ctxt = UpdatedRepresenterConfigurationContext(
                            type(self.context),
                            rpr.content_type,
                            options={XML_VALIDATE_OPTION:False})
            with ctxt:
                data = rpr.data_from_bytes(self.request.body)
        else:
            data = rpr.data_from_bytes(self.request.body)
        return data



"""
Filter specification visitors.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 5, 2011.
"""
from everest.entities.utils import slug_from_identifier
from everest.querying.base import CqlExpression
from everest.querying.base import SpecificationVisitor
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.operators import CONTAINED
from everest.querying.operators import CONTAINS
from everest.querying.operators import ENDS_WITH
from everest.querying.operators import EQUAL_TO
from everest.querying.operators import GREATER_OR_EQUALS
from everest.querying.operators import GREATER_THAN
from everest.querying.operators import IN_RANGE
from everest.querying.operators import LESS_OR_EQUALS
from everest.querying.operators import LESS_THAN
from everest.querying.operators import STARTS_WITH
from everest.resources.interfaces import IResource
from everest.resources.utils import resource_to_url
from functools import reduce as func_reduce
from operator import and_ as operator_and
from operator import or_ as operator_or
from pyramid.compat import string_types
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['CqlFilterExpression',
           'CqlFilterSpecificationVisitor',
           'FilterSpecificationVisitor',
           'RepositoryFilterSpecificationVisitor',
           ]


class FilterSpecificationVisitor(SpecificationVisitor):
    """
    Abstract base class for filter specification visitors.
    """

    def _disjunction_op(self, spec, *expressions):
        raise NotImplementedError('Abstract method.')

    def _negation_op(self, spec, expression):
        raise NotImplementedError('Abstract method.')

    def _starts_with_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _ends_with_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _contains_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _contained_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _equal_to_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _less_than_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _less_than_or_equal_to_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _greater_than_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _greater_than_or_equal_to_op(self, spec):
        raise NotImplementedError('Abstract method.')

    def _in_range_op(self, spec):
        raise NotImplementedError('Abstract method.')


class CqlFilterExpression(CqlExpression):
    """
    CQL expression representing a filter criterion.
    """
    __cql_format = '%(attr)s:%(op)s:%(val)s'
    __cql_or = ','
    __cql_not = 'not-'

    def __init__(self, attr_name, op_name, value):
        CqlExpression.__init__(self)
        self.attr_name = attr_name
        self.op_name = op_name
        self.value = value

    def _as_string(self):
        #: Returns this CQL expression as a string.
        return self.__cql_format % dict(attr=self.attr_name,
                                        op=slug_from_identifier(self.op_name),
                                        val=self.value)

    def __or__(self, other):
        if self.attr_name != other.attr_name \
           or self.op_name != other.op_name:
            raise ValueError('Attribute name and operator need to be the same '
                             'for CQL OR operation.')
        return CqlFilterExpression(self.attr_name, self.op_name,
                                   self.__cql_or.join(
                                                (self.value, other.value)))

    def __invert__(self):
        if self.op_name in [STARTS_WITH.name, ENDS_WITH.name, CONTAINED.name,
                            CONTAINS.name, EQUAL_TO.name, IN_RANGE.name]:
            op_name = "%s%s" % (self.__cql_not, self.op_name)
        elif self.op_name == LESS_THAN.name:
            op_name = GREATER_OR_EQUALS.name
        elif self.op_name == GREATER_THAN.name:
            op_name = LESS_OR_EQUALS.name
        elif self.op_name == LESS_OR_EQUALS.name:
            op_name = GREATER_THAN.name
        elif self.op_name == GREATER_OR_EQUALS.name:
            op_name = LESS_THAN.name
        else:
            raise ValueError('Invalid (non-invertible) operator name %s.'
                             % self.op_name)
        return CqlFilterExpression(self.attr_name, op_name, self.value)


@implementer(IFilterSpecificationVisitor)
class CqlFilterSpecificationVisitor(FilterSpecificationVisitor):
    """
    Filter specification visitor building a CQL expression.
    """

    __cql_range_format = '%(from_value)s-%(to_value)s'

    def _starts_with_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   STARTS_WITH.name,
                                   self.__preprocess_value(spec.attr_value))

    def _ends_with_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   ENDS_WITH.name,
                                   self.__preprocess_value(spec.attr_value))

    def _contains_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   CONTAINS.name,
                                   self.__preprocess_value(spec.attr_value))

    def _contained_op(self, spec):
        value_string = \
            ','.join([self.__preprocess_value(val) for val in spec.attr_value])
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   CONTAINED.name,
                                   value_string)

    def _equal_to_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   EQUAL_TO.name,
                                   self.__preprocess_value(spec.attr_value))

    def _less_than_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   LESS_THAN.name,
                                   self.__preprocess_value(spec.attr_value))

    def _less_than_or_equal_to_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   LESS_OR_EQUALS.name,
                                   self.__preprocess_value(spec.attr_value))

    def _greater_than_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   GREATER_THAN.name,
                                   self.__preprocess_value(spec.attr_value))

    def _greater_than_or_equal_to_op(self, spec):
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   GREATER_OR_EQUALS.name,
                                   self.__preprocess_value(spec.attr_value))

    def _in_range_op(self, spec):
        from_value, to_value = spec.attr_value
        value = self.__cql_range_format % dict(
                          from_value=self.__preprocess_value(from_value),
                          to_value=self.__preprocess_value(to_value)
                          )
        return CqlFilterExpression(self.__preprocess_attribute(spec.attr_name),
                                   IN_RANGE.name,
                                   value)

    def _conjunction_op(self, spec, *expressions):
        return func_reduce(operator_and, expressions)

    def _disjunction_op(self, spec, *expressions):
        return func_reduce(operator_or, expressions)

    def _negation_op(self, spec, expression):
        return ~expression

    def __preprocess_attribute(self, attr_name):
        return slug_from_identifier(attr_name)

    def __preprocess_value(self, value):
        if isinstance(value, string_types):
            result = '"%s"' % value
        elif IResource.providedBy(value): # pylint: disable=E1101
            result = '"%s"' % resource_to_url(value)
        else:
            result = str(value)
        return result


class RepositoryFilterSpecificationVisitor(FilterSpecificationVisitor): # pylint: disable=W0223
    """
    Specification visitors that build filter expressions for a repository
    backend.
    """
    def __init__(self, entity_class):
        FilterSpecificationVisitor.__init__(self)
        self._entity_class = entity_class

    def filter_query(self, query):
        """
        Returns the given query filtered by this visitor's filter expression.
        """
        return query.filter(self.expression)




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 14, 2014.
"""
from unittest import TestCase

from pyramid.compat import configparser
from pyramid.paster import setup_logging
from pyramid.registry import Registry
from pyramid.testing import DummyRequest
from pytest import fixture # pylint: disable=E0611
from pytest import yield_fixture # pylint: disable=E0611
import transaction

from everest.configuration import Configurator
from everest.ini import EverestIni
from everest.querying.specifications import FilterSpecificationFactory
from everest.repositories.constants import REPOSITORY_DOMAINS
from everest.repositories.interfaces import IRepositoryManager
from everest.repositories.rdb.session import ScopedSessionMaker as Session
from everest.representers.interfaces import IRepresenterRegistry
from everest.resources.interfaces import IService
from everest.testing import EverestTestApp
from everest.testing import tear_down_registry
from everest.tests.complete_app import fixtures
from paste.deploy import loadapp
from everest.querying.specifications import OrderSpecificationFactory


__docformat__ = 'reStructuredText en'
__all__ = ['app_creator',
           'configurator',
           'entity_repo',
           'ini',
           'resource_repo',
           ]


# FIXME: Is there no better way to make this available?
collection = fixtures.collection
member = fixtures.member
representer = fixtures.representer
member_representer = fixtures.member_representer
mapping = fixtures.mapping
member_mapping = fixtures.member_mapping


def pytest_addoption(parser):
    """
    This adds the `--app-ini-file` option for configuring test runs with an
    `ini` file.

    Just like when configuring pyramid with an ini file, you can not only
    define WSGI application settings in your ini file, but also set up the
    logging system if you supply a "loggers" section.
    """
    parser.addoption("--app-ini-file", action="store", default=None,
                     help="everest application ini file.")


def pytest_configure(config):
    """
    Called by pytest after all options have been collected and all plugins
    have been initialized.

    This sets up the logging system from the "loggers" section in your
    application ini file, if configured.
    """
    app_ini_file = config.getoption('--app-ini-file')
    if not app_ini_file is None:
        setup_logging(app_ini_file)


def pytest_collection_modifyitems(session, config, items): # pylint: disable=W0613
    """
    Called by pytest after all tests have been collected.

    For compatibility with the existing test suite, we remove all tests that
    were collected from abstract base test classes or from classes that have
    a `__test__` attribute set to `False` (nose feature).
    """
    removed = 0
    for idx, item in enumerate(items[:]):
        cls = item.parent.cls
        if not cls is None and issubclass(cls, TestCase) \
           and (cls.__name__.startswith('_')
                or getattr(cls, '__test__', None) is False):
            items.pop(idx - removed)
            removed += 1


@fixture(scope='session')
def ini(request):
    """
    Fixture for all tests that parse an `ini` file.

    The ini file name is pulled from the `--app-ini-file` command line
    option.
    """
    app_ini_file = request.config.getoption('--app-ini-file')
    return EverestIni(app_ini_file)


class _ConfiguratorFactory(object):
    __registries = {}

    @classmethod
    def get(cls, request, ini): # redefining ini pylint: disable=W0621
        package_name, zcml_file_name, settings = cls.__get_args(request, ini)
        return cls.__make_new(package_name, zcml_file_name, settings)

    @classmethod
    def get_cached(cls, request, ini): # redefining ini pylint: disable=W0621
        package_name, zcml_file_name, settings = cls.__get_args(request, ini)
        key = (package_name, zcml_file_name,
               _ConfiguratorFactory.__make_settings_key(settings))
        conf = _ConfiguratorFactory.__registries.get(key)
        if conf is None:
            conf = cls.__make_new(package_name, zcml_file_name, settings)
            _ConfiguratorFactory.__registries[key] = conf
        return conf

    @classmethod
    def __get_args(cls, request, ini): # redefining ini pylint: disable=W0621
        ini_section_name = getattr(request.cls, 'ini_section_name', None)
        pkg_name = getattr(request.cls, 'package_name', None)
        if not pkg_name is None:
            def_cfg_zcml = getattr(request.cls, 'config_file_name',
                                   'configure.zcml')
        else:
            def_cfg_zcml = None
        if not ini_section_name is None:
            settings = ini.get_settings(ini_section_name)
        else:
            try:
                settings = ini.get_settings('DEFAULT')
            except configparser.NoSectionError:
                settings = None
        if not settings is None:
            cfg_zcml = settings.get('configure_zcml', def_cfg_zcml)
        else:
            cfg_zcml = def_cfg_zcml
        return pkg_name, cfg_zcml, settings

    @classmethod
    def __make_new(cls, package_name, zcml_file_name, settings):
        reg = Registry('testing')
        conf = Configurator(registry=reg, package=package_name)
        conf.setup_registry(settings=settings)
        if not package_name is None:
            conf.begin()
            try:
                conf.load_zcml(zcml_file_name)
            finally:
                conf.end()
        return conf

    @classmethod
    def __make_settings_key(cls, settings):
        if settings is None:
            key = None
        else:
            key = tuple([item for item in sorted(settings.items())])
        return key


@fixture(scope='class')
def configurator(request, ini): # redefining ini pylint: disable=W0621
    """
    Fixture for all tests that set up a Pyramid configurator.
    """
    return _ConfiguratorFactory.get_cached(request, ini)


@fixture
def new_configurator(request, ini): # redefining ini pylint: disable=W0621
    """
    Fixture for all tests that set up a Pyramid configurator and modify it.
    """
    return _ConfiguratorFactory.get(request, ini)


@fixture
def mapping_registry_factory(new_configurator): # redefining new_configurator pylint: disable=W0621
    rpr_reg = new_configurator.registry.queryUtility(IRepresenterRegistry)
    return rpr_reg.get_mapping_registry


class EntityCreatorContextManager(object):
    """
    Context manager for setting up the everest framework for entity level
    operations.
    """
    def __init__(self, config):
        self.__config = config

    def __enter__(self):
        repo_mgr = self.__config.get_registered_utility(IRepositoryManager)
        repo_mgr.initialize_all()
        self.__config.begin()
        return repo_mgr.get_default()

    def __exit__(self, ext_type, value, tb):
        transaction.abort()
        repo_mgr = self.__config.get_registered_utility(IRepositoryManager)
        repo_mgr.reset_all()
        self.__config.end()


@yield_fixture
def entity_repo(configurator): # redefining configurator pylint: disable=W0621
    """
    Fixture for all tests that perform operations on entities.
    """
    with EntityCreatorContextManager(configurator) as repo:
        yield repo


class ResourceCreatorContextManager(object):
    """
    Context manager for setting up the everest framework for resource level
    operations.
    """
    def __init__(self, config, app_url, repo_name=None):
        self.__config = config
        self.__app_url = app_url
        self.__repo_name = repo_name

    def __enter__(self):
        request = DummyRequest(application_url=self.__app_url,
                               host_url=self.__app_url,
                               path_url=self.__app_url,
                               url=self.__app_url,
                               registry=self.__config.registry)
        srvc = self.__config.get_registered_utility(IService)
        self.__config.begin(request=request)
        request.root = srvc
        repo_mgr = self.__config.get_registered_utility(IRepositoryManager)
        repo_mgr.initialize_all()
        srvc.start()
        if self.__repo_name is None:
            repo = repo_mgr.get_default()
        else:
            repo = repo_mgr.get(self.__repo_name)
        return repo

    def __exit__(self, ext_type, value, tb):
        transaction.abort()
        srvc = self.__config.get_registered_utility(IService)
        srvc.stop()
        repo_mgr = self.__config.get_registered_utility(IRepositoryManager)
        repo_mgr.reset_all()
        self.__config.end()


@yield_fixture
def resource_repo(ini, configurator): # redefining ini, configurator pylint: disable=W0621
    """
    Fixture for all tests that perform operations on resources.
    """
    app_url = ini.get_app_url()
    with ResourceCreatorContextManager(configurator, app_url) as repo:
        yield repo


@yield_fixture
def system_resource_repo(ini, configurator): # redefining ini, configurator pylint: disable=W0621
    """
    Like `resource_repo`, but yields the system repository instead of the
    default repository.
    """
    app_url = ini.get_app_url()
    with ResourceCreatorContextManager(configurator, app_url,
                                       repo_name=REPOSITORY_DOMAINS.SYSTEM) \
         as repo:
        yield repo


@yield_fixture(scope='class')
def app_creator(request, ini): # redefining ini pylint: disable=W0621
    """
    Fixture for all tests that perform operations on applications.
    """
    app_name = getattr(request.cls, 'app_name', None)
    pkg_name = getattr(request.cls, 'package_name', 'everest')
    extra_environ = getattr(request.cls, 'extra_environ', {})
    with AppCreatorContextManager(ini.ini_file_path, app_name, pkg_name,
                                  extra_environ) as app:
        yield app


class AppCreatorContextManager(object):
    """
    Context manager for setting up the everest framework for application
    level (functional) operations.
    """
    def __init__(self, ini_file_path, app_name, pkg_name, extra_environ):
        wsgiapp = loadapp('config:' + ini_file_path, name=app_name)
        self.__app = EverestTestApp(wsgiapp, extra_environ=extra_environ)
        self.__config = Configurator(registry=wsgiapp.registry,
                                     package=pkg_name)

    def __enter__(self):
        self.__config.begin()
        return self.__app

    def __exit__(self, ext_type, value, tb):
        transaction.abort()
        self.__config.end()
        tear_down_registry(self.__config.registry)


# FIXME: This should be made optional (requires rdb repository).
@fixture(scope='class')
def rdb(request):
    """
    Fixture for all tests that use the relational database backend.

    This fixture has class scope.
    """
    def tear_down():
        Session.remove()
    request.addfinalizer(tear_down)


@fixture
def filter_specification_factory():
    return FilterSpecificationFactory()


@fixture
def order_specification_factory():
    return OrderSpecificationFactory()



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 22, 2011.
"""
from everest.messaging import IUserMessageNotifier
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.resources import FooMember
from everest.views.postcollection import PostCollectionView
from everest.views.putmember import PutMemberView
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['UserMessagePutMemberView',
           'UserMessagePostCollectionView',
           ]


class UserMessagePostCollectionView(PostCollectionView):
    message = 'This is my warning message'
    def _extract_request_data(self):
        foo = FooEntity(name=self.request.body)
        return FooMember(foo)

    def _process_request_data(self, data):
        reg = get_current_registry()
        msg_notifier = reg.getUtility(IUserMessageNotifier)
        if msg_notifier.notify(self.message):
            return PostCollectionView._process_request_data(self, data)


class UserMessagePutMemberView(PutMemberView):
    message = 'This is my member warning message'
    def _extract_request_data(self):
        reg = get_current_registry()
        msg_notifier = reg.getUtility(IUserMessageNotifier)
        msg_notifier.notify(self.message)
        return self.context

    def _process_request_data(self, data):
        return self._get_result(data)


class _ExtractDataExceptionViewMixin(object):
    def _extract_request_data(self):
        raise ValueError()


class _ProcessDataExceptionViewMixin(object):
    def _process_request_data(self, data): # pylint: disable=W0613
        raise ValueError()


class ExceptionPutMemberView(_ExtractDataExceptionViewMixin, PutMemberView):
    pass


class ExceptionPostCollectionView(_ProcessDataExceptionViewMixin,
                                  PutMemberView):
    pass
